# pylint: disable=redefined-outer-name

import pytest

from starknet_py.constants import ETH_FEE_CONTRACT_ADDRESS, STRK_FEE_CONTRACT_ADDRESS
from starknet_py.contract import Contract
from starknet_py.net.account.base_account import BaseAccount


@pytest.fixture(scope="package")
def eth_fee_contract(account: BaseAccount, cairo_0_fee_contract_abi) -> Contract:
    """
    Returns an instance of the ETH fee contract. It is used to transfer tokens.
    """

    return Contract(
        address=ETH_FEE_CONTRACT_ADDRESS,
        abi=cairo_0_fee_contract_abi,
        provider=account,
        cairo_version=0,
    )


@pytest.fixture(scope="package")
def strk_fee_contract(account: BaseAccount, cairo_0_fee_contract_abi) -> Contract:
    """
    Returns an instance of the STRK fee contract. It is used to transfer tokens.
    """

    return Contract(
        address=STRK_FEE_CONTRACT_ADDRESS,
        abi=cairo_0_fee_contract_abi,
        provider=account,
        cairo_version=0,
    )


@pytest.fixture(scope="package")
def cairo_0_fee_contract_abi():
    return [
        {
            "inputs": [
                {"name": "recipient", "type": "felt"},
                {"name": "amount", "type": "Uint256"},
            ],
            "name": "transfer",
            "outputs": [{"name": "success", "type": "felt"}],
            "type": "function",
        },
        {
            "members": [
                {"name": "low", "offset": 0, "type": "felt"},
                {"name": "high", "offset": 1, "type": "felt"},
            ],
            "name": "Uint256",
            "size": 2,
            "type": "struct",
        },
    ]



import pytest


@pytest.mark.asyncio
async def test_deploying_in_multicall(account, map_class_hash, map_abi):
    # pylint: disable=import-outside-toplevel,
    # docs: start
    from starknet_py.contract import Contract
    from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping
    from starknet_py.net.udc_deployer.deployer import Deployer

    # First, create Deployer instance. For more details see previous paragraph
    deployer = Deployer()

    # Create contract deployment. We will be deploying the `map` contract
    deploy_call, address = deployer.create_contract_deployment(
        class_hash=map_class_hash
    )
    # docs: end
    # docs: start

    # Address of the `map` contract is known here, so we can create its instance!
    map_contract = Contract(
        address=address, abi=map_abi, provider=account, cairo_version=1
    )

    # And now we can prepare a call
    put_call = map_contract.functions["put"].prepare_invoke_v3(key=10, value=20)

    # After that multicall transaction can be sent
    # Note that `deploy_call` and `put_call` are two regular calls!
    invoke_tx = await account.sign_invoke_v3(
        calls=[deploy_call, put_call],
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )

    resp = await account.client.send_transaction(invoke_tx)
    await account.client.wait_for_tx(resp.transaction_hash)

    (value,) = await map_contract.functions["get"].call(key=10)
    # value = 20
    # docs: end

    assert value == 20



# data from cairo repository: crates/cairo-lang-starknet/src/abi_test.rs
from collections import OrderedDict

from starknet_py.abi.v2 import Abi
from starknet_py.cairo.data_types import EventType, StructType, UintType

pool_id_struct = StructType("PoolId", OrderedDict(value=UintType(256)))

pool_id_added_event: EventType = EventType(
    "PoolIdAdded", OrderedDict(pool_id=pool_id_struct), []
)

abi_v2 = Abi(
    defined_structures={},
    events={
        "PoolIdAdded": pool_id_added_event,
    },
    functions={},
    constructor=None,
    l1_handler=None,
    defined_enums={},
    interfaces={},
    implementations={},
)



import pytest

from starknet_py.constants import ETH_FEE_CONTRACT_ADDRESS
from starknet_py.net.full_node_client import _is_valid_eth_address
from starknet_py.net.networks import default_token_address_for_network


def test_is_valid_eth_address():
    assert _is_valid_eth_address("0x333333f332a06ECB5D20D35da44ba07986D6E203")
    assert not _is_valid_eth_address("0x1")
    assert not _is_valid_eth_address("123")


def test_default_token_address_for_network():
    res = default_token_address_for_network("mainnet")
    assert res == ETH_FEE_CONTRACT_ADDRESS

    res = default_token_address_for_network("sepolia")
    assert res == ETH_FEE_CONTRACT_ADDRESS

    with pytest.raises(
        ValueError,
        match="Argument token_address must be specified when using a custom net address",
    ):
        _ = default_token_address_for_network("")



from unittest.mock import AsyncMock, patch

import pytest

from starknet_py.contract import Contract, PreparedFunctionInvokeV3
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import Call, ResourceBounds, ResourceBoundsMapping
from starknet_py.net.models import InvokeV3
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import transaction_mock_with_tip


@pytest.mark.asyncio
async def test_prepare_and_invoke_v3(map_contract):
    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(
        key=1, value=2, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    assert isinstance(prepared_invoke, PreparedFunctionInvokeV3)

    invocation = await prepared_invoke.invoke()
    assert isinstance(invocation.invoke_transaction, InvokeV3)
    assert invocation.invoke_transaction.resource_bounds == MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_invoke_v3(map_contract):
    invocation = await map_contract.functions["put"].invoke_v3(
        key=1, value=2, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    assert isinstance(invocation.invoke_transaction, InvokeV3)
    assert invocation.invoke_transaction.resource_bounds == MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_auto_fee_estimation_v3(map_contract):
    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(key=1, value=2)
    assert isinstance(prepared_invoke, PreparedFunctionInvokeV3)

    invocation = await prepared_invoke.invoke(auto_estimate=True)
    assert isinstance(invocation.invoke_transaction, InvokeV3)
    assert invocation.invoke_transaction.resource_bounds is not None


@pytest.mark.asyncio
async def test_throws_invoke_v3_without_resource_bounds(map_contract):
    error_message = (
        "One of arguments: "
        "resource_bounds or auto_estimate must be specified when invoking a transaction."
    )

    with pytest.raises(ValueError, match=error_message):
        await map_contract.functions["put"].invoke_v3(2, 3)


@pytest.mark.asyncio
async def test_throws_prepared_invoke_v3_without_resource_bounds(map_contract):
    error_message = (
        "One of arguments: "
        "resource_bounds or auto_estimate must be specified when invoking a transaction."
    )

    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(2, 3)
    assert isinstance(prepared_invoke, PreparedFunctionInvokeV3)

    with pytest.raises(ValueError, match=error_message):
        await prepared_invoke.invoke()


@pytest.mark.asyncio
async def test_throws_when_invoke_v3_with_resource_bounds_and_auto_estimate(
    map_contract,
):
    error_message = (
        "Arguments auto_estimate and resource_bounds are mutually exclusive."
    )

    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(key=2, value=3)
    with pytest.raises(ValueError, match=error_message):
        await prepared_invoke.invoke(
            resource_bounds=MAX_RESOURCE_BOUNDS, auto_estimate=True
        )


@pytest.mark.asyncio
async def test_throws_when_invoke_v3_with_tip_and_auto_estimate_tip(map_contract):
    error_message = "Arguments tip and auto_estimate_tip are mutually exclusive."

    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(key=2, value=3)
    with pytest.raises(ValueError, match=error_message):
        await prepared_invoke.invoke(
            resource_bounds=MAX_RESOURCE_BOUNDS,
            tip=12345,
            auto_estimate_tip=True,
        )


@pytest.mark.asyncio
async def test_default_tip_is_zero(map_contract):
    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(key=2, value=3)
    result = await prepared_invoke.invoke(
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    assert result.invoke_transaction.tip == 0


@pytest.mark.asyncio
async def test_provided_tip_is_used(map_contract):
    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(key=2, value=3)
    result = await prepared_invoke.invoke(
        resource_bounds=MAX_RESOURCE_BOUNDS, tip=11223344
    )
    assert result.invoke_transaction.tip == 11223344


@pytest.mark.asyncio
async def test_auto_estimate_tip_is_used(
    map_contract, get_block_with_txs_path, block_with_tips_mock
):
    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        block_with_tips_mock.transactions = [transaction_mock_with_tip(1000)]
        get_block_with_txs_mock.return_value = block_with_tips_mock

        prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(
            key=2, value=3
        )
        result = await prepared_invoke.invoke(
            resource_bounds=MAX_RESOURCE_BOUNDS,
            auto_estimate_tip=True,
        )

    assert result.invoke_transaction.tip == 1000


@pytest.mark.asyncio
async def test_latest_resource_bounds_takes_precedence(map_contract):
    prepared_function = map_contract.functions["put"].prepare_invoke_v3(
        key=1, value=2, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    invocation = await prepared_function.invoke(
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(
                max_amount=MAX_RESOURCE_BOUNDS.l1_gas.max_amount + 30,
                max_price_per_unit=MAX_RESOURCE_BOUNDS.l1_gas.max_price_per_unit + 30,
            ),
            l2_gas=ResourceBounds(
                max_amount=MAX_RESOURCE_BOUNDS.l2_gas.max_amount + 30,
                max_price_per_unit=MAX_RESOURCE_BOUNDS.l2_gas.max_price_per_unit + 30,
            ),
            l1_data_gas=ResourceBounds(
                max_amount=MAX_RESOURCE_BOUNDS.l1_data_gas.max_amount + 30,
                max_price_per_unit=MAX_RESOURCE_BOUNDS.l1_data_gas.max_price_per_unit
                + 30,
            ),
        )
    )

    assert isinstance(invocation.invoke_transaction, InvokeV3)

    for resource in ["l1_gas", "l2_gas", "l1_data_gas"]:
        for attr in ["max_amount", "max_price_per_unit"]:
            assert (
                getattr(
                    getattr(invocation.invoke_transaction.resource_bounds, resource),
                    attr,
                )
                == getattr(getattr(MAX_RESOURCE_BOUNDS, resource), attr) + 30
            )


@pytest.mark.asyncio
async def test_latest_resource_bounds_take_precedence(map_contract):
    prepared_function = map_contract.functions["put"].prepare_invoke_v3(
        key=1, value=2, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    updated_resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(
            max_amount=MAX_RESOURCE_BOUNDS.l1_gas.max_amount + 100,
            max_price_per_unit=MAX_RESOURCE_BOUNDS.l1_gas.max_price_per_unit + 200,
        ),
        l2_gas=ResourceBounds(
            max_amount=MAX_RESOURCE_BOUNDS.l2_gas.max_amount + 100,
            max_price_per_unit=MAX_RESOURCE_BOUNDS.l2_gas.max_price_per_unit + 200,
        ),
        l1_data_gas=ResourceBounds(
            max_amount=MAX_RESOURCE_BOUNDS.l1_data_gas.max_amount + 100,
            max_price_per_unit=MAX_RESOURCE_BOUNDS.l1_data_gas.max_price_per_unit + 200,
        ),
    )
    invocation = await prepared_function.invoke(resource_bounds=updated_resource_bounds)

    assert isinstance(invocation.invoke_transaction, InvokeV3)
    assert (
        invocation.invoke_transaction.resource_bounds.l1_gas
        == updated_resource_bounds.l1_gas
    )
    assert (
        invocation.invoke_transaction.resource_bounds.l2_gas
        == updated_resource_bounds.l2_gas
    )
    assert (
        invocation.invoke_transaction.resource_bounds.l1_data_gas
        == updated_resource_bounds.l1_data_gas
    )


@pytest.mark.asyncio
async def test_prepare_without_resource_bounds(map_contract):
    prepared_invoke = map_contract.functions["put"].prepare_invoke_v3(key=1, value=2)

    assert prepared_invoke.resource_bounds is None


@pytest.mark.asyncio
@pytest.mark.parametrize("key, value", ((2, 13), (412312, 32134), (12345, 3567)))
async def test_invoke_v3_and_call(key, value, map_contract):
    invocation = await map_contract.functions["put"].invoke_v3(
        key, value, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await invocation.wait_for_acceptance()
    assert isinstance(invocation.invoke_transaction, InvokeV3)

    (response,) = await map_contract.functions["get"].call(key)
    assert response == value


@pytest.mark.asyncio
async def test_call_uninitialized_contract(client):
    with pytest.raises(ClientError, match="Contract not found"):
        await client.call_contract(
            Call(
                to_addr=1,
                selector=get_selector_from_name("get_nonce"),
                calldata=[],
            ),
            block_hash="latest",
        )


@pytest.mark.asyncio
async def test_wait_for_tx(client, map_contract):
    transaction = await map_contract.functions["put"].invoke_v3(
        key=10, value=20, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await client.wait_for_tx(transaction.hash)


@pytest.mark.asyncio
async def test_error_when_prepare_without_account(client, map_contract):
    contract = await Contract.from_address(map_contract.address, client)

    with pytest.raises(
        ValueError,
        match="Contract instance was created without providing an Account.",
    ):
        contract.functions["put"].prepare_invoke_v3(key=10, value=10)


@pytest.mark.asyncio
async def test_error_when_invoke_without_account(client, map_contract):
    contract = await Contract.from_address(map_contract.address, client)

    with pytest.raises(
        ValueError,
        match="Contract instance was created without providing an Account.",
    ):
        await contract.functions["put"].invoke_v3(key=10, value=10)



import re

import pytest

from starknet_py.constants import FIELD_PRIME
from starknet_py.serialization.data_serializers.int_serializer import IntSerializer
from starknet_py.serialization.errors import InvalidTypeException, InvalidValueException

u128_serializer = IntSerializer(bits=128)

MIN_I128 = -(2**127)
MAX_I128 = 2**127 - 1


def _felt(x: int) -> int:
    if abs(x) >= FIELD_PRIME:
        raise ValueError("Value is out of field range.")

    return x + FIELD_PRIME if x < 0 else x


@pytest.mark.parametrize(
    "value, serializer, serialized_value",
    [
        (0, u128_serializer, [_felt(0)]),
        # positive values
        (1, u128_serializer, [_felt(1)]),
        (1000, u128_serializer, [_felt(1000)]),
        # negative values
        (-1, u128_serializer, [_felt(-1)]),
        (-1000, u128_serializer, [_felt(-1000)]),
        # boundaries
        (MIN_I128, u128_serializer, [_felt(MIN_I128)]),
        (MAX_I128, u128_serializer, [_felt(MAX_I128)]),
    ],
)
def test_valid_values(value, serializer, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value

    serialized = serializer.serialize(value)
    assert serialized == serialized_value


def test_deserialize_invalid_i128_values():
    error_message = re.escape(
        f"Error at path 'int128': expected value in range [{MIN_I128};{MAX_I128}]"
    )
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.deserialize([MIN_I128 - 1])
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.deserialize([MAX_I128 + 1])


def test_serialize_invalid_i128_value():
    error_message = re.escape(f"Error: expected value in range [{MIN_I128};{MAX_I128}]")
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.serialize(2**128)
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.serialize(-(2**128))


def test_invalid_type():
    error_message = re.escape(
        "Error: expected int, received 'wololoo' of type '<class 'str'>'."
    )
    with pytest.raises(InvalidTypeException, match=error_message):
        u128_serializer.serialize("wololoo")  # type: ignore



from typing import List

from starknet_py.hash.utils import message_signature
from starknet_py.net.models import AddressRepresentation, parse_address
from starknet_py.net.models.chains import ChainId
from starknet_py.net.models.transaction import AccountTransaction
from starknet_py.net.signer.base_signer import BaseSigner
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.utils.typed_data import TypedData


class StarkCurveSigner(BaseSigner):
    def __init__(
        self,
        account_address: AddressRepresentation,
        key_pair: KeyPair,
        chain_id: ChainId,
    ):
        """
        :param account_address: Address of the account contract.
        :param key_pair: Key pair of the account contract.
        :param chain_id: ChainId of the chain.
        """
        self.address = parse_address(account_address)
        self.key_pair = key_pair
        self.chain_id = chain_id

    @property
    def private_key(self) -> int:
        """Private key of the signer."""
        return self.key_pair.private_key

    @property
    def public_key(self) -> int:
        return self.key_pair.public_key

    def sign_transaction(
        self,
        transaction: AccountTransaction,
    ) -> List[int]:
        tx_hash = transaction.calculate_hash(self.chain_id)
        # pylint: disable=invalid-name
        r, s = message_signature(msg_hash=tx_hash, priv_key=self.private_key)
        return [r, s]

    def sign_message(self, typed_data: TypedData, account_address: int) -> List[int]:
        msg_hash = typed_data.message_hash(account_address)
        # pylint: disable=invalid-name
        r, s = message_signature(msg_hash=msg_hash, priv_key=self.private_key)
        return [r, s]



# pylint: disable=unused-variable
from unittest.mock import patch

import pytest

from starknet_py.constants import ETH_FEE_CONTRACT_ADDRESS
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.account import Account
from starknet_py.net.client_models import Call, ResourceBounds, ResourceBoundsMapping
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import StarknetChainId
from starknet_py.net.models.typed_data import TypedDataDict
from starknet_py.net.signer.key_pair import KeyPair


def test_init():
    # docs-start: init
    account = Account(
        address=0x123,
        client=FullNodeClient(node_url="https://your.node.url"),
        key_pair=KeyPair(12, 34),
        chain=StarknetChainId.SEPOLIA,
    )
    # docs-end: init


@pytest.mark.asyncio
async def test_execute_v3(account, contract_address):
    # docs-start: execute_v3
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
        l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    )
    resp = await account.execute_v3(
        Call(
            to_addr=contract_address,
            selector=get_selector_from_name("increase_balance"),
            calldata=[123],
        ),
        resource_bounds=resource_bounds,
    )
    # or
    # docs-end: execute_v3
    call1 = call2 = Call(
        to_addr=contract_address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[123],
    )
    # docs-start: execute_v3
    resp = await account.execute_v3(calls=[call1, call2], auto_estimate=True)
    # docs-end: execute_v3


@pytest.mark.asyncio
@patch(
    "starknet_py.net.account.account.Account.get_balance",
)
async def test_get_balance(account):
    # docs-start: get_balance
    eth_balance = await account.get_balance()
    # or with custom token contract address
    token_address = 0x1 or 1 or "0x1"
    # docs-end: get_balance
    token_address = ETH_FEE_CONTRACT_ADDRESS
    # docs-start: get_balance
    balance = await account.get_balance(token_address)
    # docs-end: get_balance


def test_sign_message(account):
    # docs-start: sign_message
    signature = account.sign_message(
        typed_data=TypedDataDict(
            types={
                "StarkNetDomain": [
                    {"name": "name", "type": "felt"},
                    {"name": "version", "type": "felt"},
                    {"name": "chainId", "type": "felt"},
                ],
                "Example": [
                    {"name": "value", "type": "felt"},
                ],
            },
            primaryType="Example",
            domain={"name": "StarkNet Example", "version": "1", "chainId": "1"},
            message={"value": 1},
        )
    )
    # docs-end: sign_message


def test_verify_message(account):
    # docs-start: verify_message
    is_correct = account.verify_message(
        typed_data=TypedDataDict(
            types={
                "StarkNetDomain": [
                    {"name": "name", "type": "felt"},
                    {"name": "version", "type": "felt"},
                    {"name": "chainId", "type": "felt"},
                ],
                "Example": [
                    {"name": "value", "type": "felt"},
                ],
            },
            primaryType="Example",
            domain={"name": "StarkNet Example", "version": "1", "chainId": "1"},
            message={"value": 1},
        ),
        signature=[12, 34],
    )
    # docs-end: verify_message



# This fixture was added to enable using async fixtures
import asyncio

import pytest


@pytest.fixture(scope="package")
def event_loop():
    policy = asyncio.get_event_loop_policy()
    loop = policy.new_event_loop()
    yield loop
    loop.close()



# pylint: disable=redefined-outer-name
from typing import Tuple

import pytest
import pytest_asyncio

from starknet_py.contract import Contract
from starknet_py.net.account.account import Account
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import DeployAccountV3
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.net.udc_deployer.deployer import Deployer
from starknet_py.tests.e2e.client.fixtures.prepare_net_for_gateway_test import (
    PreparedNetworkData,
)
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.utils import (
    _get_random_private_key_unsafe,
    _new_address,
    get_deploy_account_transaction,
    prepay_account,
)


@pytest_asyncio.fixture(scope="package")
async def deploy_account_transaction(
    account_with_validate_deploy_class_hash: int,
    eth_fee_contract: Contract,
    strk_fee_contract: Contract,
    devnet,
) -> DeployAccountV3:
    """
    Returns a DeployAccount transaction
    """
    key_pair = KeyPair.from_private_key(_get_random_private_key_unsafe())

    address, salt = _new_address(
        account_with_validate_deploy_class_hash,
        [key_pair.public_key],
    )

    await prepay_account(
        address=address,
        eth_fee_contract=eth_fee_contract,
        strk_fee_contract=strk_fee_contract,
    )

    return await get_deploy_account_transaction(
        address=address,
        key_pair=key_pair,
        class_hash=account_with_validate_deploy_class_hash,
        salt=salt,
        client=FullNodeClient(devnet),
    )


@pytest.fixture(scope="package")
def deploy_account_transaction_hash(
    prepare_network: Tuple[str, PreparedNetworkData]
) -> int:
    """
    Returns hash of deploy account transaction
    """
    _, prepared_data = prepare_network
    return prepared_data.deploy_account_transaction_hash


@pytest.fixture(scope="package")
def block_with_deploy_account_number(
    prepare_network: Tuple[str, PreparedNetworkData]
) -> int:
    """
    Returns number of the block with deploy account transaction
    """
    _, prepared_data = prepare_network
    return prepared_data.block_with_deploy_account_number


@pytest_asyncio.fixture(scope="package")
async def hello_starknet_deploy_transaction_address(
    account: Account, hello_starknet_class_hash
) -> int:
    deployer = Deployer()
    contract_deployment = deployer.create_contract_deployment_raw(
        class_hash=hello_starknet_class_hash
    )
    deploy_invoke_transaction = await account.sign_invoke_v3(
        calls=contract_deployment.call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    resp = await account.client.send_transaction(deploy_invoke_transaction)
    await account.client.wait_for_tx(resp.transaction_hash)
    return contract_deployment.address


@pytest_asyncio.fixture(scope="package")
async def block_with_declare_v3_number(hello_starknet_tx_hash: int, client) -> int:
    """
    Returns number of the block with DeclareV3 transaction
    """
    declare_v3_receipt = await client.get_transaction_receipt(hello_starknet_tx_hash)
    return declare_v3_receipt.block_number



import pytest

from starknet_py.net.account.account import Account
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import StarknetChainId
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.tests.e2e.fixtures.constants import (
    SEPOLIA_ACCOUNT_ADDRESS,
    SEPOLIA_ACCOUNT_PRIVATE_KEY,
    SEPOLIA_RPC_URL,
)


@pytest.fixture(scope="package")
def client_sepolia_testnet() -> FullNodeClient:
    return FullNodeClient(node_url=SEPOLIA_RPC_URL())


# pylint: disable=redefined-outer-name
@pytest.fixture(scope="package")
def account_sepolia_testnet(
    client_sepolia_testnet: FullNodeClient,
) -> Account:
    return Account(
        address=SEPOLIA_ACCOUNT_ADDRESS(),
        client=client_sepolia_testnet,
        key_pair=KeyPair.from_private_key(int(SEPOLIA_ACCOUNT_PRIVATE_KEY(), 0)),
        chain=StarknetChainId.SEPOLIA,
    )



from collections import OrderedDict

from starknet_py.serialization.data_serializers.array_serializer import ArraySerializer
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.data_serializers.payload_serializer import (
    PayloadSerializer,
)


def test_remove_array_lengths():
    serializer = PayloadSerializer(
        OrderedDict(
            x=FeltSerializer(),
            y=FeltSerializer(),
            values_len=FeltSerializer(),
            values=ArraySerializer(FeltSerializer()),
        )
    )
    # len parameters have to be removed
    assert set(serializer.serializers.keys()) == {"x", "y", "values"}

    data = {"x": 1, "y": 2, "values": [0, 0, 0]}
    serialized = [1, 2, 3, 0, 0, 0]
    assert serializer.serialize(data) == serialized
    assert serializer.deserialize(serialized) == tuple(data.values())



from marshmallow import fields, post_load, validate

from starknet_py.net.client_models import (
    EmittedEvent,
    EmittedEventWithFinalityStatus,
    Event,
    EventsChunk,
)
from starknet_py.net.schemas.common import Felt, FinalityStatusField
from starknet_py.utils.schema import Schema


class EventSchema(Schema):
    from_address = Felt(data_key="from_address", required=True)
    keys = fields.List(Felt(), data_key="keys", required=True)
    data = fields.List(Felt(), data_key="data", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Event:
        return Event(**data)


class EmittedEventSchema(EventSchema):
    transaction_hash = Felt(data_key="transaction_hash", required=True)
    transaction_index = fields.Integer(
        data_key="transaction_index",
        required=True,
        validate=validate.Range(min=0, error="`transaction_index` must be >= 0"),
    )
    event_index = fields.Integer(
        data_key="event_index",
        required=True,
        validate=validate.Range(min=0, error="`event_index` must be >= 0"),
    )
    block_hash = Felt(data_key="block_hash", load_default=None)
    block_number = fields.Integer(data_key="block_number", load_default=None)

    @post_load
    def make_dataclass(self, data, **kwargs) -> EmittedEvent:
        return EmittedEvent(**data)


class EmittedEventWithFinalitySchema(EmittedEventSchema):
    finality_status = FinalityStatusField(data_key="finality_status", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> EmittedEventWithFinalityStatus:
        return EmittedEventWithFinalityStatus(**data)


class EventsChunkSchema(Schema):
    events = fields.List(
        fields.Nested(EmittedEventSchema()),
        data_key="events",
        required=True,
    )
    continuation_token = fields.String(data_key="continuation_token", load_default=None)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return EventsChunk(**data)



import pytest

from starknet_py.constants import FIELD_PRIME
from starknet_py.serialization.data_serializers.array_serializer import ArraySerializer
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer

felt_array_serializer = ArraySerializer(FeltSerializer())


@pytest.mark.parametrize(
    "serializer, value, serialized_value",
    [
        (felt_array_serializer, [], [0]),
        (felt_array_serializer, [1, 2, FIELD_PRIME - 1], [3, 1, 2, FIELD_PRIME - 1]),
        # 4 nested arrays and last filled with felts
        (
            ArraySerializer(ArraySerializer(ArraySerializer(felt_array_serializer))),
            [[[[22, 38]]]],
            [1, 1, 1, 2, 22, 38],
        ),
    ],
)
def test_valid_values(serializer, value, serialized_value):
    serialized = serializer.serialize(value)
    deserialized = serializer.deserialize(serialized_value)

    assert deserialized == value
    assert serialized == serialized_value



from collections import OrderedDict, namedtuple

import pytest

from starknet_py.serialization.data_serializers.array_serializer import ArraySerializer
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.data_serializers.named_tuple_serializer import (
    NamedTupleSerializer,
)
from starknet_py.serialization.tuple_dataclass import TupleDataclass

felt_array_serializer = ArraySerializer(FeltSerializer())


# Used to generate dict, TupleDataclass and namedtuple. All of them are accepted
# by the NamedTupleSerializer.
def to_different_formats(data: TupleDataclass):
    as_dict = data.as_dict()
    yield as_dict
    yield data
    named_tuple_type = namedtuple("SomeTuple", as_dict)
    yield named_tuple_type(**as_dict)


def test_reversed_order():
    serializer = NamedTupleSerializer(
        OrderedDict(x=FeltSerializer(), y=FeltSerializer(), z=FeltSerializer())
    )
    assert serializer.serialize({"z": 3, "y": 2, "x": 1}) == [1, 2, 3]


@pytest.mark.parametrize(
    "serializer, value, serialized_value",
    [
        (
            NamedTupleSerializer(OrderedDict(x=FeltSerializer(), y=FeltSerializer())),
            TupleDataclass.from_dict({"x": 1, "y": 2}),
            [1, 2],
        ),
        (
            NamedTupleSerializer(
                OrderedDict(
                    inner=NamedTupleSerializer(
                        OrderedDict(x=FeltSerializer(), y=felt_array_serializer)
                    )
                )
            ),
            TupleDataclass.from_dict(
                {"inner": TupleDataclass.from_dict({"x": 22, "y": [38]})}
            ),
            [22, 1, 38],
        ),
    ],
)
def test_valid_values(serializer, value, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value

    for data in to_different_formats(value):
        serialized = serializer.serialize(data)
        assert serialized == serialized_value



from typing import Dict, List, Optional, Tuple, Union, cast

import aiohttp

from starknet_py.constants import RPC_CONTRACT_ERROR
from starknet_py.hash.utils import keccak256
from starknet_py.net.client import Client
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    BlockHashAndNumber,
    BlockStateUpdate,
    BlockTransactionTrace,
    Call,
    ContractsStorageKeys,
    DeclareTransactionResponse,
    DeployAccountTransactionResponse,
    DeprecatedContractClass,
    EstimatedFee,
    EventsChunk,
    Hash,
    L1HandlerTransaction,
    LatestTag,
    MessageStatus,
    PreConfirmedBlockStateUpdate,
    PreConfirmedStarknetBlock,
    PreConfirmedStarknetBlockWithReceipts,
    PreConfirmedStarknetBlockWithTxHashes,
    SentTransactionResponse,
    SierraContractClass,
    SimulatedTransaction,
    SimulationFlag,
    StarknetBlock,
    StarknetBlockWithReceipts,
    StarknetBlockWithTxHashes,
    StorageProofResponse,
    SyncStatus,
    Tag,
    Transaction,
    TransactionReceiptWithBlockInfo,
    TransactionStatusResponse,
    TransactionTrace,
)
from starknet_py.net.client_utils import (
    _create_broadcasted_txn,
    _get_raw_block_identifier,
    _is_valid_eth_address,
    _to_rpc_felt,
    _to_storage_key,
    encode_l1_message,
    get_block_identifier,
)
from starknet_py.net.executable_models import CasmClass
from starknet_py.net.http_client import RpcHttpClient
from starknet_py.net.models.transaction import (
    AccountTransaction,
    DeclareV3,
    DeployAccountV3,
    InvokeV3,
)
from starknet_py.net.schemas.contracts_storage_keys import ContractsStorageKeysSchema
from starknet_py.net.schemas.rpc.block import (
    BlockHashAndNumberSchema,
    BlockStateUpdateSchema,
    PreConfirmedBlockStateUpdateSchema,
    PreConfirmedStarknetBlockSchema,
    PreConfirmedStarknetBlockWithReceiptsSchema,
    PreConfirmedStarknetBlockWithTxHashesSchema,
    StarknetBlockSchema,
    StarknetBlockWithReceiptsSchema,
    StarknetBlockWithTxHashesSchema,
)
from starknet_py.net.schemas.rpc.contract import (
    CasmClassSchema,
    DeprecatedContractClassSchema,
    SierraContractClassSchema,
    SyncStatusSchema,
)
from starknet_py.net.schemas.rpc.event import EventsChunkSchema
from starknet_py.net.schemas.rpc.general import EstimatedFeeSchema
from starknet_py.net.schemas.rpc.storage_proof import StorageProofResponseSchema
from starknet_py.net.schemas.rpc.trace_api import (
    BlockTransactionTraceSchema,
    SimulatedTransactionSchema,
    TransactionTraceSchema,
)
from starknet_py.net.schemas.rpc.transactions import (
    DeclareTransactionResponseSchema,
    DeployAccountTransactionResponseSchema,
    MessageStatusSchema,
    SentTransactionSchema,
    TransactionReceiptWithBlockInfoSchema,
    TransactionStatusResponseSchema,
    TypesOfTransactionsSchema,
)
from starknet_py.transaction_errors import TransactionNotReceivedError
from starknet_py.utils.sync import add_sync_methods


@add_sync_methods
class FullNodeClient(Client):
    # pylint: disable=too-many-public-methods
    def __init__(
        self,
        node_url: str,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        Client for interacting with Starknet json-rpc interface.

        :param node_url: Url of the node providing rpc interface
        :param session: Aiohttp session to be used for request. If not provided, client will create a session for
                        every request. When using a custom session, user is responsible for closing it manually.
        """
        self.url = node_url
        self._client = RpcHttpClient(url=node_url, session=session)

    async def get_block(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlock, PreConfirmedStarknetBlock]:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getBlockWithTxs",
            params=block_identifier,
        )

        if block_identifier == {"block_id": "pre_confirmed"}:
            return cast(
                PreConfirmedStarknetBlock, PreConfirmedStarknetBlockSchema().load(res)
            )
        return cast(StarknetBlock, StarknetBlockSchema().load(res))

    async def get_block_with_txs(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlock, PreConfirmedStarknetBlock]:
        return await self.get_block(block_hash=block_hash, block_number=block_number)

    async def get_block_with_tx_hashes(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlockWithTxHashes, PreConfirmedStarknetBlockWithTxHashes]:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getBlockWithTxHashes",
            params=block_identifier,
        )

        if block_identifier == {"block_id": "pre_confirmed"}:
            return cast(
                PreConfirmedStarknetBlockWithTxHashes,
                PreConfirmedStarknetBlockWithTxHashesSchema().load(res),
            )
        return cast(
            StarknetBlockWithTxHashes,
            StarknetBlockWithTxHashesSchema().load(res),
        )

    async def get_block_with_receipts(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlockWithReceipts, PreConfirmedStarknetBlockWithReceipts]:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getBlockWithReceipts",
            params=block_identifier,
        )

        if block_identifier == {"block_id": "pre_confirmed"}:
            return cast(
                PreConfirmedStarknetBlockWithReceipts,
                PreConfirmedStarknetBlockWithReceiptsSchema().load(res),
            )
        return cast(
            StarknetBlockWithReceipts,
            StarknetBlockWithReceiptsSchema().load(res),
        )

    # TODO (#809): add tests with multiple emitted keys
    async def get_events(
        self,
        address: Optional[Hash] = None,
        keys: Optional[List[List[Hash]]] = None,
        *,
        from_block_number: Optional[Union[int, Tag]] = None,
        from_block_hash: Optional[Union[Hash, Tag]] = None,
        to_block_number: Optional[Union[int, Tag]] = None,
        to_block_hash: Optional[Union[Hash, Tag]] = None,
        follow_continuation_token: bool = False,
        continuation_token: Optional[str] = None,
        chunk_size: int = 1,
    ) -> EventsChunk:
        # pylint: disable=too-many-arguments
        """
        :param address: The address of the contract that emitted the event.
        :param keys: List consisting lists of keys by which the events are filtered. They match the keys *by position*,
            e.g. given an event with 3 keys, [[1,2],[],[3]] which should return events that have either 1 or 2 in
            the first key, any value for their second key and 3 for their third key.
        :param from_block_number: Number of the block from which events searched for **starts**
            or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`. Mutually exclusive
            with ``from_block_hash`` parameter.
            If not provided, query starts from block 0.
        :param from_block_hash: Hash of the block from which events searched for **starts**
            or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`. Mutually exclusive with
            ``from_block_number`` parameter.
            If not provided, query starts from block 0.
        :param to_block_number: Number of the block to which events searched for **end**
            or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`. Mutually exclusive with
            ``to_block_hash`` parameter.
            If not provided, query ends at block `"pre_confirmed"`.
        :param to_block_hash: Hash of the block to which events searched for **end**
            or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`. Mutually exclusive with
            ``to_block_number`` parameter.
            If not provided, query ends at block `"pre_confirmed"`.
        :param follow_continuation_token: Flag deciding whether all events should be collected during one function call,
            defaults to False.
        :param continuation_token: Continuation token from which the returned events start.
        :param chunk_size: Size of chunk of events returned by one ``get_events`` call, defaults to 1 (minimum).

        :return: ``EventsResponse`` dataclass containing events and optional continuation token.
        """

        if chunk_size <= 0:
            raise ValueError("Argument chunk_size must be greater than 0.")

        if keys is None:
            keys = []
        if address is not None:
            address = _to_rpc_felt(address)
        if from_block_number is None and from_block_hash is None:
            from_block_number = 0

        from_block = _get_raw_block_identifier(from_block_hash, from_block_number)
        to_block = _get_raw_block_identifier(to_block_hash, to_block_number)
        keys = [[_to_rpc_felt(key) for key in inner_list] for inner_list in keys]

        events_list = []
        while True:
            events, continuation_token = await self._get_events_chunk(
                from_block=from_block,
                to_block=to_block,
                address=address,
                keys=keys,
                chunk_size=chunk_size,
                continuation_token=continuation_token,
            )
            events_list.extend(events)
            if not follow_continuation_token or continuation_token is None:
                break

        events_response = cast(
            EventsChunk,
            EventsChunkSchema().load(
                {"events": events_list, "continuation_token": continuation_token}
            ),
        )

        return events_response

    async def _get_events_chunk(
        self,
        from_block: Union[dict, Hash, Tag, None],
        to_block: Union[dict, Hash, Tag, None],
        keys: List[List[Hash]],
        chunk_size: int,
        address: Optional[Hash] = None,
        continuation_token: Optional[str] = None,
    ) -> Tuple[list, Optional[str]]:
        # pylint: disable=too-many-arguments
        params = {
            "chunk_size": chunk_size,
            "from_block": from_block,
            "to_block": to_block,
            "keys": keys,
        }
        if continuation_token is not None:
            params["continuation_token"] = continuation_token
        if address is not None:
            params["address"] = address

        res = await self._client.call(
            method_name="getEvents",
            params={"filter": params},
        )

        if "continuation_token" in res:
            return res["events"], res["continuation_token"]
        return res["events"], None

    async def get_state_update(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[BlockStateUpdate, PreConfirmedBlockStateUpdate]:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getStateUpdate",
            params=block_identifier,
        )

        if block_identifier == {"block_id": "pre_confirmed"}:
            return cast(
                PreConfirmedBlockStateUpdate,
                PreConfirmedBlockStateUpdateSchema().load(res),
            )
        return cast(BlockStateUpdate, BlockStateUpdateSchema().load(res))

    async def get_storage_at(
        self,
        contract_address: Hash,
        key: int,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getStorageAt",
            params={
                "contract_address": _to_rpc_felt(contract_address),
                "key": _to_storage_key(key),
                **block_identifier,
            },
        )
        res = cast(str, res)
        return int(res, 16)

    async def get_storage_proof(
        self,
        block_hash: Optional[Union[Hash, LatestTag]] = None,
        block_number: Optional[Union[int, LatestTag]] = None,
        class_hashes: Optional[List[int]] = None,
        contract_addresses: Optional[List[int]] = None,
        contracts_storage_keys: Optional[List[ContractsStorageKeys]] = None,
    ) -> StorageProofResponse:
        if block_hash == "pre_confirmed" or block_number == "pre_confirmed":
            raise ValueError(
                "'pre_confirmed' block tag is not allowed gor `get_storage_proof`"
            )

        class_hashes_serialized = (
            [_to_rpc_felt(class_hash) for class_hash in class_hashes]
            if class_hashes
            else []
        )
        contract_addresses_serialized = (
            [_to_rpc_felt(contract_address) for contract_address in contract_addresses]
            if contract_addresses
            else []
        )
        contracts_storage_keys_serialized = (
            (
                [
                    cast(
                        Dict,
                        ContractsStorageKeysSchema().dump(obj=key),
                    )
                    for key in contracts_storage_keys
                ]
            )
            if contracts_storage_keys
            else []
        )

        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number, allow_pre_confirmed=False
        )

        params = {
            "class_hashes": class_hashes_serialized,
            "contract_addresses": contract_addresses_serialized,
            "contracts_storage_keys": contracts_storage_keys_serialized,
            **block_identifier,
        }

        res = await self._client.call(
            method_name="getStorageProof",
            params=params,
        )
        return cast(StorageProofResponse, StorageProofResponseSchema().load(res))

    async def get_transaction(
        self,
        tx_hash: Hash,
    ) -> Transaction:
        try:
            res = await self._client.call(
                method_name="getTransactionByHash",
                params={"transaction_hash": _to_rpc_felt(tx_hash)},
            )
        except ClientError as ex:
            raise TransactionNotReceivedError() from ex

        return cast(Transaction, TypesOfTransactionsSchema().load(res))

    async def get_l1_message_hash(self, tx_hash: Hash) -> Hash:
        """
        :param tx_hash: Transaction's hash
        :return: Message hash
        """
        tx = await self.get_transaction(tx_hash)
        if not isinstance(tx, L1HandlerTransaction):
            raise TypeError(
                f"Transaction {tx_hash} is not a result of L1->L2 interaction."
            )

        encoded_message = encode_l1_message(tx)
        return keccak256(encoded_message)

    async def get_transaction_receipt(
        self, tx_hash: Hash
    ) -> TransactionReceiptWithBlockInfo:
        res = await self._client.call(
            method_name="getTransactionReceipt",
            params={"transaction_hash": _to_rpc_felt(tx_hash)},
        )

        return cast(
            TransactionReceiptWithBlockInfo,
            TransactionReceiptWithBlockInfoSchema().load(res),
        )

    async def estimate_fee(
        self,
        tx: Union[AccountTransaction, List[AccountTransaction]],
        skip_validate: bool = False,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[EstimatedFee, List[EstimatedFee]]:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        if single_transaction := isinstance(tx, AccountTransaction):
            tx = [tx]

        res = await self._client.call(
            method_name="estimateFee",
            params={
                "request": [_create_broadcasted_txn(transaction=t) for t in tx],
                "simulation_flags": (
                    [SimulationFlag.SKIP_VALIDATE] if skip_validate else []
                ),
                **block_identifier,
            },
        )

        if single_transaction:
            res = res[0]

        return cast(
            EstimatedFee,
            EstimatedFeeSchema().load(res, many=not single_transaction),
        )

    async def estimate_message_fee(
        self,
        from_address: str,
        to_address: Hash,
        entry_point_selector: Hash,
        payload: List[Hash],
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> EstimatedFee:
        # pylint: disable=too-many-arguments
        """
        :param from_address: The address of the L1 (Ethereum) contract sending the message.
        :param to_address: The target L2 (Starknet) address the message is sent to.
        :param entry_point_selector: The selector of the l1_handler in invoke in the target contract.
        :param payload: Payload of the message.
        :param block_hash: Hash of the requested block or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`.
            Mutually exclusive with ``block_number`` parameter. If not provided, queries block `"pre_confirmed"`.
        :param block_number: Number (height) of the requested block or literals
            `"l1_accepted"`, `"pre_confirmed"` or `"latest"`.
            Mutually exclusive with ``block_hash`` parameter. If not provided, queries block `"pre_confirmed"`.
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        assert _is_valid_eth_address(
            from_address
        ), f"Argument 'from_address': {from_address} is not a valid Ethereum address."

        message_body = {
            "from_address": from_address,
            "to_address": _to_rpc_felt(to_address),
            "entry_point_selector": _to_rpc_felt(entry_point_selector),
            "payload": [_to_rpc_felt(x) for x in payload],
        }

        try:
            res = await self._client.call(
                method_name="estimateMessageFee",
                params={
                    "message": message_body,
                    **block_identifier,
                },
            )
            return cast(EstimatedFee, EstimatedFeeSchema().load(res))
        except ClientError as err:
            if err.code == RPC_CONTRACT_ERROR:
                raise ClientError(
                    err.message
                    + f" Note that your ETH address ('from_address': {from_address}) might be invalid"
                ) from err
            raise err

    async def get_block_number(self) -> int:
        """Get the most recent accepted block number"""
        return await self._client.call(method_name="blockNumber", params={})

    async def get_block_hash_and_number(self) -> BlockHashAndNumber:
        """Get the most recent accepted block hash and number"""
        res = await self._client.call(method_name="blockHashAndNumber", params={})
        return cast(BlockHashAndNumber, BlockHashAndNumberSchema().load(res))

    async def get_chain_id(self) -> str:
        return await self._client.call(method_name="chainId", params={})

    async def get_messages_status(self, transaction_hash: Hash) -> List[MessageStatus]:
        res = await self._client.call(
            method_name="getMessagesStatus",
            params={"transaction_hash": _to_rpc_felt(transaction_hash)},
        )
        return cast(
            List[MessageStatus],
            MessageStatusSchema().load(res, many=True),
        )

    async def get_syncing_status(self) -> Union[bool, SyncStatus]:
        """Returns an object about the sync status, or false if the node is not syncing"""
        sync_status = await self._client.call(method_name="syncing", params={})
        if isinstance(sync_status, bool):
            return sync_status
        return cast(SyncStatus, SyncStatusSchema().load(sync_status))

    async def call_contract(
        self,
        call: Call,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> List[int]:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )
        res = await self._client.call(
            method_name="call",
            params={
                "request": {
                    "contract_address": _to_rpc_felt(call.to_addr),
                    "entry_point_selector": _to_rpc_felt(call.selector),
                    "calldata": [_to_rpc_felt(i1) for i1 in call.calldata],
                },
                **block_identifier,
            },
        )
        return [int(i, 16) for i in res]

    async def send_transaction(self, transaction: InvokeV3) -> SentTransactionResponse:
        params = _create_broadcasted_txn(transaction=transaction)

        res = await self._client.call(
            method_name="addInvokeTransaction",
            params={"invoke_transaction": params},
        )

        return cast(SentTransactionResponse, SentTransactionSchema().load(res))

    async def deploy_account(
        self, transaction: DeployAccountV3
    ) -> DeployAccountTransactionResponse:
        params = _create_broadcasted_txn(transaction=transaction)

        res = await self._client.call(
            method_name="addDeployAccountTransaction",
            params={"deploy_account_transaction": params},
        )

        return cast(
            DeployAccountTransactionResponse,
            DeployAccountTransactionResponseSchema().load(res),
        )

    async def declare(self, transaction: DeclareV3) -> DeclareTransactionResponse:
        params = _create_broadcasted_txn(transaction=transaction)

        res = await self._client.call(
            method_name="addDeclareTransaction",
            params={"declare_transaction": {**params}},
        )

        return cast(
            DeclareTransactionResponse,
            DeclareTransactionResponseSchema().load(res),
        )

    async def get_class_hash_at(
        self,
        contract_address: Hash,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )
        res = await self._client.call(
            method_name="getClassHashAt",
            params={
                "contract_address": _to_rpc_felt(contract_address),
                **block_identifier,
            },
        )
        res = cast(str, res)
        return int(res, 16)

    async def get_class_by_hash(
        self,
        class_hash: Hash,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[SierraContractClass, DeprecatedContractClass]:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getClass",
            params={
                "class_hash": _to_rpc_felt(class_hash),
                **block_identifier,
            },
        )

        if "sierra_program" in res:
            return cast(
                SierraContractClass,
                SierraContractClassSchema().load(res),
            )
        return cast(DeprecatedContractClass, DeprecatedContractClassSchema().load(res))

    async def get_transaction_by_block_id(
        self,
        index: int,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Transaction:
        """
        Get the details of transaction in block identified by block_hash and transaction index.

        :param index: Index of the transaction
        :param block_hash: Hash of the block
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: Transaction object
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getTransactionByBlockIdAndIndex",
            params={
                **block_identifier,
                "index": index,
            },
        )
        return cast(Transaction, TypesOfTransactionsSchema().load(res))

    async def get_block_transaction_count(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        Get the number of transactions in a block given a block id

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: Number of transactions in the designated block
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getBlockTransactionCount",
            params=block_identifier,
        )
        res = cast(int, res)
        return res

    async def get_class_at(
        self,
        contract_address: Hash,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[SierraContractClass, DeprecatedContractClass]:
        """
        Get the contract class definition in the given block at the given address

        :param contract_address: The address of the contract whose class definition will be returned
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: Contract declared to Starknet
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        res = await self._client.call(
            method_name="getClassAt",
            params={
                **block_identifier,
                "contract_address": _to_rpc_felt(contract_address),
            },
        )

        if "sierra_program" in res:
            return cast(
                SierraContractClass,
                SierraContractClassSchema().load(res),
            )
        return cast(DeprecatedContractClass, DeprecatedContractClassSchema().load(res))

    async def get_contract_nonce(
        self,
        contract_address: Hash,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )
        res = await self._client.call(
            method_name="getNonce",
            params={
                "contract_address": _to_rpc_felt(contract_address),
                **block_identifier,
            },
        )
        res = cast(str, res)
        return int(res, 16)

    async def get_compiled_casm(self, class_hash: int) -> CasmClass:
        res = await self._client.call(
            method_name="getCompiledCasm",
            params={"class_hash": _to_rpc_felt(class_hash)},
        )
        return cast(CasmClass, CasmClassSchema().load(res))

    async def spec_version(self) -> str:
        """
        Returns the version of the Starknet JSON-RPC specification being used.

        :return: String with version of the Starknet JSON-RPC specification.
        """
        spec_version = cast(
            str,
            await self._client.call(
                method_name="specVersion",
                params={},
            ),
        )
        return spec_version

    async def get_transaction_status(self, tx_hash: Hash) -> TransactionStatusResponse:
        res = await self._client.call(
            method_name="getTransactionStatus",
            params={"transaction_hash": _to_rpc_felt(tx_hash)},
        )
        return cast(
            TransactionStatusResponse,
            TransactionStatusResponseSchema().load(res),
        )

    # ------------------------------- Trace API -------------------------------

    async def trace_transaction(
        self,
        tx_hash: Hash,
    ) -> TransactionTrace:
        """
        For a given executed transaction, returns the trace of its execution, including internal calls.

        :param tx_hash: Hash of the executed transaction.
        :return: Trace of the transaction.
        """
        res = await self._client.call(
            method_name="traceTransaction",
            params={
                "transaction_hash": _to_rpc_felt(tx_hash),
            },
        )
        return cast(TransactionTrace, TransactionTraceSchema().load(res))

    async def simulate_transactions(
        self,
        transactions: List[AccountTransaction],
        skip_validate: bool = False,
        skip_fee_charge: bool = False,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> List[SimulatedTransaction]:
        # pylint: disable=too-many-arguments
        """
        Simulates a given sequence of transactions on the requested state, and generates the execution traces.
        Note the following:

        - A transaction may revert. If this occurs, no error is thrown. Instead, revert details are visible
          in the returned trace object.
        - If a transaction reverts, this will be reflected by the revert_error property in the trace.
        - Other types of failures (e.g. unexpected error or failure in the validation phase) will result
          in TRANSACTION_EXECUTION_ERROR.

        :param transactions: Transactions to be traced.
        :param skip_validate: Flag checking whether the validation part of the transaction should be executed.
        :param skip_fee_charge: Flag deciding whether fee should be deducted from the balance before the simulation
            of the next transaction.
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: The execution trace and consumed resources for each transaction.
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number
        )

        simulation_flags = []
        if skip_validate:
            simulation_flags.append(SimulationFlag.SKIP_VALIDATE)
        if skip_fee_charge:
            simulation_flags.append(SimulationFlag.SKIP_FEE_CHARGE)

        res = await self._client.call(
            method_name="simulateTransactions",
            params={
                **block_identifier,
                "simulation_flags": simulation_flags,
                "transactions": [
                    _create_broadcasted_txn(transaction=t) for t in transactions
                ],
            },
        )

        return cast(
            List[SimulatedTransaction],
            SimulatedTransactionSchema().load(res, many=True),
        )

    async def trace_block_transactions(
        self,
        block_hash: Optional[Union[Hash, LatestTag]] = None,
        block_number: Optional[Union[int, LatestTag]] = None,
    ) -> List[BlockTransactionTrace]:
        """
        Retrieve traces for all transactions in the given block.

        :param block_hash: Block's hash or literals `"pre_confirmed"`.
        :param block_number: Block's number or literals `"pre_confirmed"`.
        :return: List of execution traces of all transactions included in the given block with transaction hashes.
        """
        block_identifier = get_block_identifier(
            block_hash=block_hash, block_number=block_number, allow_pre_confirmed=False
        )

        res = await self._client.call(
            method_name="traceBlockTransactions",
            params={
                **block_identifier,
            },
        )
        return cast(
            List[BlockTransactionTrace],
            BlockTransactionTraceSchema().load(res, many=True),
        )



from typing import Union

from starknet_py.constants import QUERY_VERSION_BASE


def _extract_tx_version(version: Union[int, str]) -> str:
    if isinstance(version, str):
        version = int(version, 16)
    return str(version % QUERY_VERSION_BASE)



import pytest

from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1560)")
async def test_signing_fee_estimate(account, map_contract):
    # docs: start
    # Create a transaction
    call = map_contract.functions["put"].prepare_invoke_v3(key=10, value=20)
    transaction = await account.sign_invoke_v3(
        calls=call,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    # Re-sign a transaction for fee estimation
    estimate_transaction = await account.sign_for_fee_estimate(transaction)

    # Transaction uses a version that cannot be executed on Starknet
    assert estimate_transaction.version == 3 + 2**128
    assert estimate_transaction.signature != transaction.signature

    # Get a fee estimation
    estimate = await account.client.estimate_fee(transaction)
    assert estimate.overall_fee > 0
    print(estimate.overall_fee)
    print(estimate.to_resource_bounds())
    # Use a new fee in original transaction
    transaction = await account.sign_invoke_v3(
        calls=call, resource_bounds=estimate.to_resource_bounds()
    )

    # Send a transaction
    result = await account.client.send_transaction(transaction)
    await account.client.wait_for_tx(result.transaction_hash)
    # docs: end



import re
from typing import List, Optional, Union, cast

from aiohttp import ClientSession

from starknet_py.devnet_utils.devnet_client_models import (
    BalanceRecord,
    Config,
    IncreaseTimeResponse,
    MintResponse,
    PostmanFlushResponse,
    PredeployedAccount,
    SetTimeResponse,
)
from starknet_py.devnet_utils.devnet_rpc_schema import (
    BalanceRecordSchema,
    ConfigSchema,
    IncreasedTimeResponseSchema,
    MintResponseSchema,
    PostmanFlushResponseSchema,
    PredeployedAccountSchema,
    SetTimeResponseSchema,
)
from starknet_py.net.client_models import Hash, PriceUnit, Tag
from starknet_py.net.full_node_client import (
    FullNodeClient,
    _get_raw_block_identifier,
    _to_rpc_felt,
)
from starknet_py.net.http_client import RpcHttpClient
from starknet_py.utils.sync import add_sync_methods


@add_sync_methods
class DevnetClient(FullNodeClient):
    def __init__(
        self,
        node_url: str = "http://127.0.0.1:5050",
        session: Optional[ClientSession] = None,
    ):
        """
        Client for interacting with Starknet devnet json-rpc interface.

        Based on https://0xspaceshard.github.io/starknet-devnet-rs/docs/intro

        :param node_url: Url of the node providing rpc interface
        :param session: Aiohttp session to be used for request. If not provided, client will create a session for
                        every request. When using a custom session, user is responsible for closing it manually.
        """

        super().__init__(node_url=node_url, session=session)
        self._devnet_client = RpcHttpClient(
            url=node_url, session=session, method_prefix="devnet"
        )

    async def impersonate_account(self, address: Hash):
        """
        Impersonate the given account.
        For impersonation to work, Devnet needs to be run in forking mode.

        :param address: Address of the account contract.
        """

        await self._devnet_client.call(
            method_name="impersonateAccount",
            params={"account_address": _to_rpc_felt(address)},
        )

    async def stop_impersonate_account(self, address: Hash):
        """
        Stop impersonating the given account.

        :param address: Address of the account contract.
        """

        await self._devnet_client.call(
            method_name="stopImpersonateAccount",
            params={"account_address": _to_rpc_felt(address)},
        )

    async def auto_impersonate(self):
        """
        Enables automatic account impersonation.
        Every account that does not exist in the local state will be impersonated.
        For impersonation to work, Devnet needs to be run in forking mode.
        """

        await self._devnet_client.call(method_name="autoImpersonate")

    async def stop_auto_impersonate(self):
        await self._devnet_client.call(method_name="stopAutoImpersonate")

    async def mint(
        self, address: Hash, amount: int, unit: Union[PriceUnit, str] = PriceUnit.FRI
    ) -> MintResponse:
        """
        Mint tokens to the given address.

        :param address: Address of the account contract.
        :param amount: Amount of tokens to mint. Must be integer.
        :param unit: Literals `"FRI"` or `"WEI"`, default to `"FRI"`.
        """

        res = await self._devnet_client.call(
            method_name="mint",
            params={
                "address": _to_rpc_felt(address),
                "amount": amount,
                "unit": unit.upper() if isinstance(unit, str) else unit.value,
            },
        )

        return cast(MintResponse, MintResponseSchema().load(res))

    async def get_account_balance(
        self,
        address: Hash,
        unit: Union[PriceUnit, str] = PriceUnit.WEI,
        block_tag: str = "latest",
    ) -> BalanceRecord:
        """
        Get the balance of the given account.

        :param address: Address of the account contract.
        :param unit: Literals `"FRI"` or `"WEI"` defaults to `"WEI"`.
        :param block_tag: Literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`, defaults to `"latest"`.
        """

        res = await self._devnet_client.call(
            method_name="getAccountBalance",
            params={
                "address": _to_rpc_felt(address),
                "unit": unit.upper() if isinstance(unit, str) else unit.value,
                "block_tag": block_tag,
            },
        )

        return cast(BalanceRecord, BalanceRecordSchema().load(res))

    async def create_block(self) -> str:
        """
        Create a new block.
        """

        res = await self._devnet_client.call(method_name="createBlock")

        return res["block_hash"]

    async def abort_block(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> List[str]:
        """
        This functionality allows simulating block abortion that can occur on mainnet.
        It is supported in the `--state-archive-capacity full` mode.

        :param block_number: Number of the block which the state of Devnet will be reverted to
            or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`.
        :param block_hash: Hash of the block which the state of Devnet will be reverted to
            or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        """

        res = await self._devnet_client.call(
            method_name="abortBlocks",
            params={
                "starting_block_id": _get_raw_block_identifier(block_hash, block_number)
            },
        )
        return res["aborted"]

    async def dump(self, path: str):
        """
        Dump the state of the devnet to a file.
        Dumping on request requires providing `--dump-on` mode on the startup.

        :param path: Path to the file.
        """

        await self._devnet_client.call(
            method_name="dump",
            params={"path": path},
        )

    async def load(self, path: str):
        """
        Load the state of the devnet from a file.

        :param path: Path to the file.
        """

        await self._devnet_client.call(
            method_name="load",
            params={"path": path},
        )

    async def restart(self):
        """
        Restart the devnet.
        """

        await self._devnet_client.call(
            method_name="restart",
        )

    async def postman_load(
        self, network_url: str, address: Optional[str] = None
    ) -> str:
        """
        Loads a `MockStarknetMessaging
        <https://github.com/0xSpaceShard/starknet-devnet-rs/blob/138120b355c44ae60269167b326d1a267f7af0a8/contracts/l1-l2-messaging/solidity/src/MockStarknetMessaging.sol>`_
        contract. The address parameter is optional; if provided, the MockStarknetMessaging contract will be fetched
        from that address, otherwise a new one will be deployed.

        :param network_url: is the URL of the JSON-RPC API of the L1 node you've run locally or that already exists

        :return: The address of the messaging contract.
        """

        params = {"network_url": network_url}
        if address is not None:
            params["address"] = address

        res = await self._devnet_client.call(
            method_name="postmanLoad",
            params=params,
        )

        return res["messaging_contract_address"]

    async def postman_flush(self, dry_run: bool = False) -> PostmanFlushResponse:
        """
        Goes through the newly enqueued messages, sending them from L1 to L2 and from L2 to L1. Requires no body.

        :param dry_run: Optional, If `True` the result of flushing will be shown without actually triggering it.

        .. warning::
            A running L1 node is required if dry_run is not set.
        """

        res = await self._devnet_client.call(
            method_name="postmanFlush",
            params={"dry_run": dry_run},
        )
        return cast(PostmanFlushResponse, PostmanFlushResponseSchema().load(res))

    # pylint: disable=too-many-arguments
    async def send_message_to_l2(
        self,
        l2_contract_address: Hash,
        entry_point_selector: Hash,
        l1_contract_address: Hash,
        payload: List[Hash],
        nonce: Hash,
        paid_fee_on_l1: Hash,
    ) -> str:
        """
        Sending mock transactions from L1 to L2 without the need for running L1.
        Deployed L2 contract address l2_contract_address and entry_point_selector must be valid
        otherwise new block will not be created. Normally nonce is calculated by L1 StarknetContract and
        it's used in L1 and L2. In this case, we need to provide it manually.

        A running L1 node is not required for this operation.

        :param l2_contract_address: Address of the L2 contract.
        :param entry_point_selector: Selector of the entry point.
        :param l1_contract_address: Address of the L1 contract.
        :param payload: List of felts.
        :param nonce: Nonce.
        :param paid_fee_on_l1: Paid fee on L1.

        :return: Transaction hash.
        """
        res = await self._devnet_client.call(
            method_name="postmanSendMessageToL2",
            params={
                "l2_contract_address": _to_rpc_felt(l2_contract_address),
                "entry_point_selector": _to_rpc_felt(entry_point_selector),
                "l1_contract_address": _to_eth_address(l1_contract_address),
                "payload": [_to_rpc_felt(entry) for entry in payload],
                "nonce": _to_rpc_felt(nonce),
                "paid_fee_on_l1": _to_rpc_felt(paid_fee_on_l1),
            },
        )

        return res["transaction_hash"]

    async def consume_message_from_l2(
        self, from_address: Hash, to_address: Hash, payload: List[Hash]
    ) -> str:
        """
        Sending mock transactions from L2 to L1. Deployed L2 contract address l2_contract_address and
        l1_contract_address must be valid.

        :param from_address: Address of the L2 contract.
        :param to_address: Address of the L1 contract.
        :param payload: List of felts.

        :return: Message hash.

        .. warning::
            A running L1 node is required for this operation.

        """
        res = await self._devnet_client.call(
            method_name="postmanConsumeMessageFromL2",
            params={
                "from_address": _to_rpc_felt(from_address),
                "to_address": _to_eth_address(to_address),
                "payload": [_to_rpc_felt(entry) for entry in payload],
            },
        )

        return res["message_hash"]

    async def get_predeployed_accounts(
        self, with_balance: bool = False
    ) -> List[PredeployedAccount]:
        """
        Get the predeployed accounts.

        :param with_balance: If `True` the balance of the accounts will be included, default to False.
        """

        res = await self._devnet_client.call(
            method_name="getPredeployedAccounts", params={"with_balance": with_balance}
        )

        return cast(
            List[PredeployedAccount], PredeployedAccountSchema().load(res, many=True)
        )

    async def get_config(self) -> Config:
        """
        Get the devnet configuration.
        """

        res = await self._devnet_client.call(method_name="getConfig")

        return cast(Config, ConfigSchema().load(res))

    async def increase_time(self, time: int) -> IncreaseTimeResponse:
        """
        Increases the block timestamp by the provided amount and generates a new block.
        All subsequent blocks will keep this increment.

        :param time: Time to increase in seconds.
        """

        res = await self._devnet_client.call(
            method_name="increaseTime", params={"time": time}
        )

        return cast(IncreaseTimeResponse, IncreasedTimeResponseSchema().load(res))

    async def set_time(
        self, time: int, generate_block: bool = False
    ) -> SetTimeResponse:
        """
        Set the time of the devnet.
        Warning: block time can be set in the past and lead to unexpected behaviour!

        :param time: Time to set in seconds. (Unix time)
        :param generate_block: If `True` a new block will be generated, default to False.
        """

        res = await self._devnet_client.call(
            method_name="setTime",
            params={"time": time, "generate_block": generate_block},
        )

        return cast(SetTimeResponse, SetTimeResponseSchema().load(res))


def _to_eth_address(value: Hash) -> str:
    """
    Convert the value to Ethereum address matching a ``^0x[a-fA-F0-9]{40}$`` pattern.

    :param value: The value to convert.
    :return: Ethereum address representation of the value.
    """
    if isinstance(value, str):
        value = int(value, 16)

    eth_address = hex(value)
    assert re.match("^0x[a-fA-F0-9]{40}$", eth_address)
    return eth_address



from dataclasses import dataclass
from typing import Generator, TypedDict, Union

from starknet_py.cairo.felt import uint256_range_check
from starknet_py.serialization._context import (
    Context,
    DeserializationContext,
    SerializationContext,
)
from starknet_py.serialization.data_serializers.cairo_data_serializer import (
    CairoDataSerializer,
)


class Uint256Dict(TypedDict):
    low: int
    high: int


@dataclass
class UintSerializer(CairoDataSerializer[Union[int, Uint256Dict], int]):
    """
    Serializer of uint. In Cairo there are few uints (u8, ..., u128 and u256).
    u256 is represented by structure {low: u128, high: u128}.
    Can serialize an int and dict.
    Deserializes data to an int.

    Examples:
        if bits < 256:
            0 => [0]
            1 => [1]
            2**128-1 => [2**128-1]
        else:
            0 => [0,0]
            1 => [1,0]
            2**128 => [0,1]
            3 + 2**128 => [3,1]
    """

    bits: int

    def deserialize_with_context(self, context: DeserializationContext) -> int:
        if self.bits < 256:
            (uint,) = context.reader.read(1)
            with context.push_entity("uint" + str(self.bits)):
                self._ensure_valid_uint(uint, context, self.bits)

            return uint

        [low, high] = context.reader.read(2)

        # Checking if resulting value is in [0, 2**256) range is not enough. Uint256 should be made of two uint128.
        with context.push_entity("low"):
            self._ensure_valid_uint(low, context, bits=128)
        with context.push_entity("high"):
            self._ensure_valid_uint(high, context, bits=128)

        return (high << 128) + low

    def serialize_with_context(
        self, context: SerializationContext, value: Union[int, Uint256Dict]
    ) -> Generator[int, None, None]:
        context.ensure_valid_type(value, isinstance(value, (int, dict)), "int or dict")
        if isinstance(value, int):
            yield from self._serialize_from_int(value, context, self.bits)
        else:
            yield from self._serialize_from_dict(context, value)

    @staticmethod
    def _serialize_from_int(
        value: int, context: SerializationContext, bits: int
    ) -> Generator[int, None, None]:
        if bits < 256:
            UintSerializer._ensure_valid_uint(value, context, bits)

            yield value
        else:
            uint256_range_check(value)

            result = (value % 2**128, value >> 128)
            yield from result

    def _serialize_from_dict(
        self, context: SerializationContext, value: Uint256Dict
    ) -> Generator[int, None, None]:
        with context.push_entity("low"):
            self._ensure_valid_uint(value["low"], context, bits=128)
            yield value["low"]
        with context.push_entity("high"):
            self._ensure_valid_uint(value["high"], context, bits=128)
            yield value["high"]

    @staticmethod
    def _ensure_valid_uint(value: int, context: Context, bits: int):
        """
        Ensures that value is a valid uint on `bits` bits.
        """
        context.ensure_valid_value(
            0 <= value < 2**bits, "expected value in range [0;2**" + str(bits) + ")"
        )



import pytest

from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping


@pytest.mark.skip(reason="Test require eth node running.")
@pytest.mark.asyncio
async def test_postman_load(devnet_client, l1_l2_contract, account):
    # pylint: disable=import-outside-toplevel
    # pylint: disable=unused-variable

    eth_account_address = 1390849295786071768276380950238675083608645509734

    # docs: start
    from starknet_py.devnet_utils.devnet_client import DevnetClient

    client = DevnetClient(node_url="http://127.0.0.1:5050")

    # docs: end
    client: DevnetClient = devnet_client
    # docs: start
    # Deploying the messaging contract on ETH network
    # e.g. anvil eth devnet https://github.com/foundry-rs/foundry/tree/master/crates/anvil
    await client.postman_load(network_url="http://127.0.0.1:8545")
    # docs: end

    # docs: messaging-contract-start
    from starknet_py.contract import Contract

    # Address of your contract that is emitting messages
    contract_address = "0x12345"

    # docs: messaging-contract-end
    contract_address = l1_l2_contract.address

    # docs: messaging-contract-start
    contract = await Contract.from_address(address=contract_address, provider=account)

    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=50000, max_price_per_unit=int(1e12)),
        l2_gas=ResourceBounds.init_with_zeros(),
        l1_data_gas=ResourceBounds.init_with_zeros(),
    )
    await contract.functions["increase_balance"].invoke_v3(
        user=account.address,
        amount=100,
        resource_bounds=resource_bounds,
    )

    # docs: messaging-contract-end
    assert await contract.functions["get_balance"].call(user=account.address) == (100,)

    # docs: messaging-contract-start
    # Invoking function that is emitting message
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=50000, max_price_per_unit=int(1e12)),
        l2_gas=ResourceBounds.init_with_zeros(),
        l1_data_gas=ResourceBounds.init_with_zeros(),
    )
    await contract.functions["withdraw"].invoke_v3(
        user=account.address,
        amount=100,
        l1_address=eth_account_address,
        resource_bounds=resource_bounds,
    )
    # docs: messaging-contract-end
    assert await contract.functions["get_balance"].call(user=account.address) == (0,)

    # docs: flush-1-start
    # Sending messages from L2 to L1.
    flush_response = await client.postman_flush()

    message = flush_response.messages_to_l1[0]

    message_hash = await client.consume_message_from_l2(
        from_address=message.from_address,
        to_address=message.to_address,
        payload=message.payload,
    )
    # docs: flush-1-end

    # docs: send-l2-start
    await client.send_message_to_l2(
        l2_contract_address=contract_address,
        entry_point_selector=get_selector_from_name("deposit"),
        l1_contract_address="0xa000000000000000000000000000000000000001",
        payload=[account.address, 100],
        nonce="0x0",
        paid_fee_on_l1="0xfffffffffff",
    )

    # Sending messages from L1 to L2.
    flush_response = await client.postman_flush()
    # docs: send-l2-end

    assert await contract.functions["get_balance"].call(user=account.address) == (100,)



import re

import pytest

from starknet_py.serialization.data_serializers.uint_serializer import UintSerializer
from starknet_py.serialization.errors import InvalidTypeException, InvalidValueException

u128_serializer = UintSerializer(bits=128)
u256_serializer = UintSerializer(bits=256)

SHIFT = 2**128
MAX_U128 = SHIFT - 1


@pytest.mark.parametrize(
    "value, serializer, serialized_value",
    [
        (123 + 456 * SHIFT, u256_serializer, [123, 456]),
        (
            21323213211421424142 + 347932774343 * SHIFT,
            u256_serializer,
            [21323213211421424142, 347932774343],
        ),
        (0, u256_serializer, [0, 0]),
        (MAX_U128, u256_serializer, [MAX_U128, 0]),
        (MAX_U128 * SHIFT, u256_serializer, [0, MAX_U128]),
        (MAX_U128 + MAX_U128 * SHIFT, u256_serializer, [MAX_U128, MAX_U128]),
        (123, u128_serializer, [123]),
        (0, u128_serializer, [0]),
        (MAX_U128, u128_serializer, [MAX_U128]),
    ],
)
def test_valid_values(value, serializer, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value

    serialized = serializer.serialize(value)
    assert serialized == serialized_value

    if serializer.bits == 256:
        assert serialized_value == serializer.serialize(
            {"low": serialized_value[0], "high": serialized_value[1]}
        )


@pytest.mark.parametrize(
    "value, uint256_part",
    [
        ([MAX_U128 + 1, 0], "low"),
        ([MAX_U128 + 1, MAX_U128 + 1], "low"),
        ([-1, 0], "low"),
        ([0, MAX_U128 + 1], "high"),
        ([0, -1], "high"),
    ],
)
def test_deserialize_invalid_256_values(value, uint256_part):
    # We need to escape braces
    error_message = re.escape(
        "Error at path '" + uint256_part + "': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=error_message):
        u256_serializer.deserialize(value)


def test_deserialize_invalid_128_values():
    # We need to escape braces
    error_message = re.escape(
        "Error at path 'uint128': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.deserialize([MAX_U128 + 1])
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.deserialize([-1])


def test_serialize_invalid_256_int_value():
    error_message = re.escape("Error: Uint256 is expected to be in range [0;2**256)")
    with pytest.raises(InvalidValueException, match=error_message):
        u256_serializer.serialize(2**256)
    with pytest.raises(InvalidValueException, match=error_message):
        u256_serializer.serialize(-1)


def test_serialize_invalid_128_int_value():
    error_message = re.escape("Error: expected value in range [0;2**128)")
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.serialize(2**128)
    with pytest.raises(InvalidValueException, match=error_message):
        u128_serializer.serialize(-1)


def test_serialize_invalid_dict_values():
    low_error_message = re.escape(
        "Error at path 'low': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=low_error_message):
        u256_serializer.serialize({"low": -1, "high": 12324})
    with pytest.raises(InvalidValueException, match=low_error_message):
        u256_serializer.serialize({"low": MAX_U128 + 1, "high": 4543535})

    high_error_message = re.escape(
        "Error at path 'high': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=high_error_message):
        u256_serializer.serialize({"low": 652432, "high": -1})
    with pytest.raises(InvalidValueException, match=high_error_message):
        u256_serializer.serialize({"low": 0, "high": MAX_U128 + 1})


@pytest.mark.parametrize("serializer", (u128_serializer, u256_serializer))
def test_invalid_type(serializer):
    error_message = re.escape(
        "Error: expected int or dict, received 'wololoo' of type '<class 'str'>'."
    )
    with pytest.raises(InvalidTypeException, match=error_message):
        serializer.serialize("wololoo")  # type: ignore



"""
Dataclasses representing responses from Starknet.
They need to stay backwards compatible for old transactions/blocks to be fetchable.

If you encounter a ValidationError in the context of an RPC response, it is possible to disable validation.
This can be achieved by setting the environment variable, STARKNET_PY_MARSHMALLOW_UNKNOWN_EXCLUDE,
to true. Consequently, any unknown fields in response will be excluded.
"""

import datetime
import json
from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Iterable, List, Literal, Optional, Union, cast

from marshmallow import EXCLUDE

from starknet_py.abi.v0.shape import AbiDictList
from starknet_py.abi.v1.schemas import (
    ContractAbiEntrySchema as ContractAbiEntrySchemaV1,
)
from starknet_py.abi.v1.shape import AbiDictEntry as AbiDictEntryV1
from starknet_py.abi.v1.shape import AbiDictList as AbiDictListV1
from starknet_py.abi.v2.schemas import (
    ContractAbiEntrySchema as ContractAbiEntrySchemaV2,
)
from starknet_py.abi.v2.shape import AbiDictEntry as AbiDictEntryV2
from starknet_py.abi.v2.shape import AbiDictList as AbiDictListV2
from starknet_py.utils.constructor_args_translator import _is_abi_v2

# pylint: disable=too-many-lines

Hash = Union[int, str]
Tag = Literal["l1_accepted", "pre_confirmed", "latest"]
LatestTag = Literal["latest"]


@dataclass
class Call:
    """
    Dataclass representing a call to Starknet contract.
    """

    to_addr: int
    selector: int
    calldata: List[int]


Calls = Union[Call, Iterable[Call]]


@dataclass
class Event:
    """
    Dataclass representing a Starknet event.
    """

    from_address: int
    keys: List[int]
    data: List[int]


@dataclass
class _EmittedEventBase(Event):
    transaction_hash: int
    transaction_index: int
    event_index: int


@dataclass
class _EmittedEventDefaultBase(Event):
    block_hash: Optional[int] = None
    block_number: Optional[int] = None


@dataclass
class EmittedEvent(_EmittedEventDefaultBase, _EmittedEventBase):
    """
    Dataclass representing an event emitted by transaction.
    """


@dataclass
class EventsChunk:
    """
    Dataclass representing events returned by FullNodeClient.get_events method.
    """

    events: List[EmittedEvent]
    continuation_token: Optional[str] = None


@dataclass
class L2toL1Message:
    """
    Dataclass representing a L2->L1 message.
    """

    payload: List[int]
    l2_address: int  # from_address in spec
    l1_address: int  # to_address in spec


@dataclass
class ResourcePrice:
    """
    Dataclass representing prices of L1 gas.
    """

    price_in_wei: int
    price_in_fri: int


@dataclass
class ResourceBounds:
    """
    Dataclass representing max amount and price of the resource that can be used in the transaction.
    """

    max_amount: int
    max_price_per_unit: int

    @staticmethod
    def init_with_zeros():
        return ResourceBounds(max_amount=0, max_price_per_unit=0)


@dataclass
class OutsideExecutionTimeBounds:
    """
    Dataclass representing time bounds within which outside execution
    transaction is valid and allowed to be executed.
    """

    execute_after: datetime.datetime
    execute_before: datetime.datetime

    @property
    def execute_after_timestamp(self) -> int:
        return int(self.execute_after.timestamp())

    @property
    def execute_before_timestamp(self) -> int:
        return int(self.execute_before.timestamp())


@dataclass
class ResourceBoundsMapping:
    """
    Dataclass representing resource limits that can be used in the transaction.
    """

    l1_gas: ResourceBounds
    l1_data_gas: ResourceBounds
    l2_gas: ResourceBounds

    @staticmethod
    def init_with_zeros():
        return ResourceBoundsMapping(
            l1_gas=ResourceBounds.init_with_zeros(),
            l1_data_gas=ResourceBounds.init_with_zeros(),
            l2_gas=ResourceBounds.init_with_zeros(),
        )


class PriceUnit(Enum):
    """
    Enum representing price unit types.
    """

    WEI = "WEI"
    FRI = "FRI"


@dataclass
class FeePayment:
    """
    Dataclass representing fee payment info as it appears in receipts.
    """

    amount: int
    unit: PriceUnit


class DAMode(Enum):
    """
    Specifies a storage domain in Starknet. Each domain has different guarantees regarding availability.
    """

    L1 = 0
    L2 = 1


class L1DAMode(Enum):
    BLOB = "BLOB"
    CALLDATA = "CALLDATA"


class TransactionType(Enum):
    """
    Enum representing transaction types.
    """

    INVOKE = "INVOKE"
    DECLARE = "DECLARE"
    DEPLOY_ACCOUNT = "DEPLOY_ACCOUNT"
    DEPLOY = "DEPLOY"
    L1_HANDLER = "L1_HANDLER"


@dataclass
class Transaction(ABC):
    """
    Dataclass representing common attributes of all transactions.
    """

    # Technically, RPC specification moved 'transaction_hash' out of the TXN object, but since it is always returned
    # together with the rest of the data, it remains here (but is still Optional just in case as spec says)
    hash: Optional[int]
    signature: List[int]
    version: int

    def __post_init__(self):
        if self.__class__ == Transaction:
            raise TypeError("Cannot instantiate abstract Transaction class.")


@dataclass
class DeprecatedTransaction(Transaction):
    """
    Dataclass representing common attributes of transactions v1 and v2.
    """

    max_fee: int

    def __post_init__(self):
        if self.__class__ == DeprecatedTransaction:
            raise TypeError("Cannot instantiate abstract DeprecatedTransaction class.")


@dataclass
class TransactionV3(Transaction):
    """
    Dataclass representing common attributes of all transactions v3.
    """

    resource_bounds: ResourceBoundsMapping
    paymaster_data: List[int]
    tip: int
    nonce_data_availability_mode: DAMode
    fee_data_availability_mode: DAMode

    def __post_init__(self):
        if self.__class__ == TransactionV3:
            raise TypeError("Cannot instantiate abstract TransactionV3 class.")


@dataclass
class InvokeTransactionV0(DeprecatedTransaction):
    """
    Dataclass representing invoke transaction v0.
    """

    calldata: List[int]
    contract_address: int
    entry_point_selector: int


@dataclass
class InvokeTransactionV1(DeprecatedTransaction):
    """
    Dataclass representing invoke transaction v1.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use `starknet_py.net.client_models.InvokeTransactionV3` instead.
    """

    calldata: List[int]
    sender_address: int
    nonce: int


@dataclass
class InvokeTransactionV3(TransactionV3):
    """
    Dataclass representing invoke transaction v3.
    """

    calldata: List[int]
    sender_address: int
    nonce: int
    account_deployment_data: List[int]


@dataclass
class DeclareTransactionV0(DeprecatedTransaction):
    """
    Dataclass representing declare transaction v0.
    """

    sender_address: int
    class_hash: int


@dataclass
class DeclareTransactionV1(DeprecatedTransaction):
    """
    Dataclass representing declare transaction v1.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use `starknet_py.net.client_models.DeclareTransactionV3` instead.
    """

    sender_address: int
    class_hash: int
    nonce: int


@dataclass
class DeclareTransactionV2(DeprecatedTransaction):
    """
    Dataclass representing declare transaction v2.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use `starknet_py.net.client_models.DeclareTransactionV3` instead.
    """

    sender_address: int
    class_hash: int
    compiled_class_hash: int
    nonce: int


@dataclass
class DeclareTransactionV3(TransactionV3):
    """
    Dataclass representing declare transaction v3.
    """

    sender_address: int
    class_hash: int
    compiled_class_hash: int
    nonce: int
    account_deployment_data: List[int]


@dataclass
class DeployTransaction(Transaction):
    """
    Dataclass representing deploy transaction.
    """

    contract_address_salt: int
    constructor_calldata: List[int]
    class_hash: int


@dataclass
class DeployAccountTransactionV1(DeprecatedTransaction):
    """
    Dataclass representing deploy account transaction v1.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use `starknet_py.net.client_models.DeployAccountTransactionV3` instead.
    """

    nonce: int
    contract_address_salt: int
    constructor_calldata: List[int]
    class_hash: int


@dataclass
class DeployAccountTransactionV3(TransactionV3):
    """
    Dataclass representing deploy account transaction v3.
    """

    nonce: int
    contract_address_salt: int
    constructor_calldata: List[int]
    class_hash: int


@dataclass
class L1HandlerTransaction(Transaction):
    """
    Dataclass representing l1 handler transaction.
    """

    contract_address: int
    calldata: List[int]
    entry_point_selector: int
    nonce: int


class TransactionStatus(Enum):
    """
    Enum representing transaction statuses.
    """

    RECEIVED = "RECEIVED"
    CANDIDATE = "CANDIDATE"
    PRE_CONFIRMED = "PRE_CONFIRMED"
    ACCEPTED_ON_L2 = "ACCEPTED_ON_L2"
    ACCEPTED_ON_L1 = "ACCEPTED_ON_L1"


class TransactionStatusWithoutL1(Enum):
    """
    Enum representing transaction statuses.
    """

    RECEIVED = "RECEIVED"
    CANDIDATE = "CANDIDATE"
    PRE_CONFIRMED = "PRE_CONFIRMED"
    ACCEPTED_ON_L2 = "ACCEPTED_ON_L2"


class TransactionExecutionStatus(Enum):
    """
    Enum representing transaction execution statuses.
    """

    SUCCEEDED = "SUCCEEDED"
    REVERTED = "REVERTED"


class TransactionFinalityStatus(Enum):
    """
    Enum representing transaction finality statuses.
    """

    PRE_CONFIRMED = "PRE_CONFIRMED"
    ACCEPTED_ON_L2 = "ACCEPTED_ON_L2"
    ACCEPTED_ON_L1 = "ACCEPTED_ON_L1"


class TransactionFinalityStatusWithoutL1(Enum):
    """
    Enum representing transaction finality statuses, without ACCEPTED_ON_L1 status.
    """

    PRE_CONFIRMED = "PRE_CONFIRMED"
    ACCEPTED_ON_L2 = "ACCEPTED_ON_L2"


@dataclass
class InnerCallExecutionResources:
    """
    Dataclass representing the resource consumed by an inner call (does not account for state diffs).
    """

    l1_gas: int
    l2_gas: int


@dataclass
class ExecutionResources:
    """
    Dataclass representing the resources consumed by the transaction, includes both computation and data.
    """

    l1_gas: int
    l1_data_gas: int
    l2_gas: int


# TODO (#1219): split into PendingTransactionReceipt and TransactionReceipt
@dataclass
class TransactionReceipt:
    """
    Dataclass representing details of sent transaction.
    """

    # pylint: disable=too-many-instance-attributes

    transaction_hash: int
    execution_status: TransactionExecutionStatus
    finality_status: TransactionFinalityStatus
    execution_resources: ExecutionResources
    actual_fee: FeePayment
    type: TransactionType

    events: List[Event] = field(default_factory=list)
    messages_sent: List[L2toL1Message] = field(default_factory=list)

    contract_address: Optional[int] = None  # DEPLOY_ACCOUNT_TXN_RECEIPT only
    message_hash: Optional[int] = None  # L1_HANDLER_TXN_RECEIPT only

    revert_reason: Optional[str] = None


@dataclass
class TransactionReceiptWithBlockInfo(TransactionReceipt):
    """
    Dataclass representing details of sent transaction with additional block info.
    """

    block_number: int = 0
    block_hash: Optional[int] = None


@dataclass
class TransactionWithReceipt:
    transaction: Transaction
    receipt: TransactionReceipt


@dataclass
class SentTransactionResponse:
    """
    Dataclass representing a result of sending a transaction to Starknet.
    """

    transaction_hash: int
    code: Optional[str] = None


@dataclass
class DeclareTransactionResponse(SentTransactionResponse):
    """
    Dataclass representing a result of declaring a contract on Starknet.
    """

    class_hash: int = 0


@dataclass
class DeployAccountTransactionResponse(SentTransactionResponse):
    """
    Dataclass representing a result of deploying an account contract to Starknet.
    """

    address: int = 0


class BlockStatus(Enum):
    """
    Enum representing block status.
    """

    PRE_CONFIRMED = "PRE_CONFIRMED"
    ACCEPTED_ON_L2 = "ACCEPTED_ON_L2"
    ACCEPTED_ON_L1 = "ACCEPTED_ON_L1"


@dataclass
class PreConfirmedBlockHeader:
    # pylint: disable=too-many-instance-attributes
    block_number: int
    timestamp: int
    sequencer_address: int
    l1_gas_price: ResourcePrice
    l2_gas_price: ResourcePrice
    l1_data_gas_price: ResourcePrice
    l1_da_mode: L1DAMode
    starknet_version: str


@dataclass
class PreConfirmedStarknetBlock(PreConfirmedBlockHeader):
    """
    Dataclass representing a pre-confirmed block on Starknet.
    """

    transactions: List[Transaction]


@dataclass
class PreConfirmedStarknetBlockWithTxHashes(PreConfirmedBlockHeader):
    """
    Dataclass representing a pre_confirmed block on Starknet containing transaction hashes.
    """

    transactions: List[int]


@dataclass
class PreConfirmedStarknetBlockWithReceipts(PreConfirmedBlockHeader):
    """
    Dataclass representing a pre_confirmed block on Starknet with txs and receipts result
    """

    transactions: List[TransactionWithReceipt]


@dataclass
class BlockHeader:
    """
    Dataclass representing a block header.
    """

    # pylint: disable=too-many-instance-attributes

    block_hash: int
    parent_hash: int
    block_number: int
    new_root: int
    timestamp: int
    sequencer_address: int
    l1_gas_price: ResourcePrice
    l2_gas_price: ResourcePrice
    l1_data_gas_price: ResourcePrice
    l1_da_mode: L1DAMode
    starknet_version: str


@dataclass
class StarknetBlock(BlockHeader):
    """
    Dataclass representing a block on Starknet.
    """

    status: BlockStatus
    transactions: List[Transaction]


@dataclass
class StarknetBlockWithTxHashes(BlockHeader):
    """
    Dataclass representing a block on Starknet containing transaction hashes.
    """

    status: BlockStatus
    transactions: List[int]


@dataclass
class StarknetBlockWithReceipts(BlockHeader):
    """
    Dataclass representing a block on Starknet with txs and receipts result
    """

    status: BlockStatus
    transactions: List[TransactionWithReceipt]


@dataclass
class BlockHashAndNumber:
    block_hash: int
    block_number: int


@dataclass
class SyncStatus:
    starting_block_hash: int
    starting_block_num: int
    current_block_hash: int
    current_block_num: int
    highest_block_hash: int
    highest_block_num: int


@dataclass
class StorageEntry:
    """
    Dataclass representing single change in the storage.
    """

    key: int
    value: int


@dataclass
class StorageDiffItem:
    """
    Dataclass representing all storage changes for the contract.
    """

    address: int
    storage_entries: List[StorageEntry]


@dataclass
class EstimatedFee:
    # pylint: disable=too-many-instance-attributes
    """
    Dataclass representing estimated fee.
    """

    l1_gas_consumed: int
    l1_gas_price: int
    l2_gas_consumed: int
    l2_gas_price: int
    l1_data_gas_consumed: int
    l1_data_gas_price: int
    overall_fee: int
    unit: PriceUnit

    def to_resource_bounds(
        self, amount_multiplier=1.5, unit_price_multiplier=1.5
    ) -> ResourceBoundsMapping:
        """
        Converts estimated fee to resource bounds with applied multipliers.

        :param amount_multiplier: Multiplier for max amount, defaults to 1.5.
        :param unit_price_multiplier: Multiplier for max price per unit, defaults to 1.5.
        :return: Resource bounds with applied multipliers.
        """

        l1_resource_bounds = ResourceBounds(
            max_amount=int(self.l1_gas_consumed * amount_multiplier),
            max_price_per_unit=int(self.l1_gas_price * unit_price_multiplier),
        )

        l2_resource_bounds = ResourceBounds(
            max_amount=int(self.l2_gas_consumed * amount_multiplier),
            max_price_per_unit=int(self.l2_gas_price * unit_price_multiplier),
        )

        l1_data_resource_bounds = ResourceBounds(
            max_amount=int(self.l1_data_gas_consumed * amount_multiplier),
            max_price_per_unit=int(self.l1_data_gas_price * unit_price_multiplier),
        )

        return ResourceBoundsMapping(
            l1_gas=l1_resource_bounds,
            l2_gas=l2_resource_bounds,
            l1_data_gas=l1_data_resource_bounds,
        )

    def calculate_overall_fee(self):
        return (
            self.l1_gas_consumed * self.l1_gas_price
            + self.l2_gas_consumed * self.l2_gas_price
            + self.l1_data_gas_consumed * self.l1_data_gas_price
        )


@dataclass
class DeployedContract:
    """
    Dataclass representing basic data of the deployed contract.
    """

    address: int
    class_hash: int


@dataclass
class ContractsNonce:
    """
    Dataclass representing nonce of the contract.
    """

    contract_address: int
    nonce: int


@dataclass
class DeclaredContractHash:
    """
    Dataclass containing hashes of the declared contract.
    """

    class_hash: int
    compiled_class_hash: int


@dataclass
class ReplacedClass:
    """
    Dataclass representing new class_hash of the contract.
    """

    contract_address: int
    class_hash: int


@dataclass
class MigratedClass:
    """
    Dataclass representing migrated compiled class.
    """

    class_hash: int
    compiled_class_hash: int


@dataclass
class StateDiff:
    """
    Dataclass representing state changes in the block.
    """

    storage_diffs: List[StorageDiffItem]
    deprecated_declared_classes: List[int]
    declared_classes: List[DeclaredContractHash]
    migrated_compiled_classes: List[MigratedClass]
    deployed_contracts: List[DeployedContract]
    replaced_classes: List[ReplacedClass]
    nonces: List[ContractsNonce]


@dataclass
class BlockStateUpdate:
    """
    Dataclass representing a change in state of a block.
    """

    block_hash: int
    new_root: int
    old_root: int
    state_diff: StateDiff


@dataclass
class PreConfirmedBlockStateUpdate:
    """
    Dataclass representing a pre_confirmed change in state of a block.
    """

    old_root: Optional[int]
    state_diff: StateDiff


@dataclass
class EntryPoint:
    """
    Dataclass representing contract entry point.
    """

    offset: int
    selector: int


@dataclass
class EntryPointsByType:
    """
    Dataclass representing contract class entrypoints by entry point type.
    """

    constructor: List[EntryPoint]
    external: List[EntryPoint]
    l1_handler: List[EntryPoint]


@dataclass
class _DeprecatedContract:
    """
    Dataclass representing contract declared on Starknet.
    """

    program: dict
    entry_points_by_type: EntryPointsByType


@dataclass
class DeprecatedContractClass(_DeprecatedContract):
    """
    Dataclass representing contract declared on Starknet.
    """

    abi: Optional[AbiDictList] = None


@dataclass
class DeprecatedCompiledContract(_DeprecatedContract):
    """
    Dataclass representing ContractClass with required abi.
    """

    # abi is a required key in CompiledContractSchema,
    # default_factory is used, since abi in ContractClass is Optional
    # and otherwise, non-keyword arguments would follow keyword arguments
    abi: AbiDictList = field(default_factory=list)

    def convert_to_deprecated_contract_class(self) -> DeprecatedContractClass:
        """
        Converts an instance of DeprecatedCompiledContract to DeprecatedContractClass.
        """
        return DeprecatedContractClass(
            program=self.program,
            entry_points_by_type=self.entry_points_by_type,
            abi=self.abi,
        )


@dataclass
class SierraEntryPoint:
    """
    Dataclass representing contract entry point
    """

    function_idx: int
    selector: int


@dataclass
class SierraEntryPointsByType:
    """
    Dataclass representing contract class entrypoints by entry point type
    """

    constructor: List[SierraEntryPoint]
    external: List[SierraEntryPoint]
    l1_handler: List[SierraEntryPoint]


@dataclass
class _SierraContract:

    contract_class_version: str
    sierra_program: List[int]
    entry_points_by_type: SierraEntryPointsByType


@dataclass
class SierraContractClass(_SierraContract):
    """
    Dataclass representing Cairo1 contract declared on Starknet
    """

    abi: Optional[str] = None

    @property
    def parsed_abi(self) -> Union[AbiDictListV2, AbiDictListV1]:
        if self.abi is None:
            return []

        load_abi: List = json.loads(self.abi)

        if _is_abi_v2(load_abi):
            return [
                cast(
                    AbiDictEntryV2,
                    ContractAbiEntrySchemaV2(unknown=EXCLUDE).load(entry),
                )
                for entry in load_abi
            ]

        return [
            cast(AbiDictEntryV1, ContractAbiEntrySchemaV1(unknown=EXCLUDE).load(entry))
            for entry in load_abi
        ]


@dataclass
class SierraCompiledContract(_SierraContract):
    """
    Dataclass representing SierraContractClass with required abi.
    """

    abi: str = field(default_factory=str)

    @property
    def parsed_abi(self) -> Union[AbiDictListV2, AbiDictListV1]:
        load_abi: List = json.loads(self.abi)

        if _is_abi_v2(load_abi):
            return [
                cast(
                    AbiDictEntryV2,
                    ContractAbiEntrySchemaV2(unknown=EXCLUDE).load(entry),
                )
                for entry in load_abi
            ]

        return [
            cast(AbiDictEntryV1, ContractAbiEntrySchemaV1(unknown=EXCLUDE).load(entry))
            for entry in load_abi
        ]

    def convert_to_sierra_contract_class(self) -> SierraContractClass:
        """
        Converts an instance of SierraCompiledContract to SierraContractClass.
        """
        return SierraContractClass(
            contract_class_version=self.contract_class_version,
            sierra_program=self.sierra_program,
            entry_points_by_type=self.entry_points_by_type,
            abi=self.abi,
        )


@dataclass
class CasmClassEntryPoint:
    """
    Dataclass representing CasmClass entrypoint.
    """

    selector: int
    offset: int
    builtins: List[str]


@dataclass
class CasmClassEntryPointsByType:
    """
    Dataclass representing CasmClass entrypoints by entry point type.
    """

    constructor: List[CasmClassEntryPoint]
    external: List[CasmClassEntryPoint]
    l1_handler: List[CasmClassEntryPoint]


@dataclass
class TransactionStatusResponse:
    """
    Dataclass representing transaction status for the FullNodeClient.
    """

    finality_status: TransactionStatus
    execution_status: Optional[TransactionExecutionStatus] = None
    failure_reason: Optional[str] = None


# ------------------------------- Trace API dataclasses -------------------------------


@dataclass
class OrderedEvent:
    """
    Dataclass representing an event alongside its order within the transaction.
    """

    keys: List[int]
    data: List[int]
    order: int


@dataclass
class OrderedMessage:
    """
    Dataclass representing a message alongside its order within the transaction.
    """

    payload: List[int]
    l2_address: int  # from_address in spec
    l1_address: int  # to_address in spec
    order: int


class SimulationFlag(str, Enum):
    """
    Enum class representing possible simulation flags for trace API.
    """

    SKIP_VALIDATE = "SKIP_VALIDATE"
    SKIP_FEE_CHARGE = "SKIP_FEE_CHARGE"


class EntryPointType(Enum):
    """
    Enum class representing entry point types.
    """

    EXTERNAL = "EXTERNAL"
    L1_HANDLER = "L1_HANDLER"
    CONSTRUCTOR = "CONSTRUCTOR"


class CallType(Enum):
    """
    Enum class representing call types.
    """

    DELEGATE = "DELEGATE"
    LIBRARY_CALL = "LIBRARY_CALL"
    CALL = "CALL"


@dataclass
class FunctionInvocation:
    """
    Dataclass representing an invocation of a function.
    """

    # pylint: disable=too-many-instance-attributes
    contract_address: int
    entry_point_selector: int
    calldata: List[int]
    caller_address: int
    class_hash: int
    entry_point_type: EntryPointType
    call_type: CallType
    result: List[int]
    calls: List["FunctionInvocation"]
    events: List[OrderedEvent]
    messages: List[OrderedMessage]
    execution_resources: InnerCallExecutionResources
    is_reverted: bool


@dataclass
class RevertedFunctionInvocation:
    """
    Dataclass representing revert reason for the transaction.
    """

    revert_reason: str


@dataclass
class InvokeTransactionTrace:
    """
    Dataclass representing a transaction trace of an INVOKE transaction.
    """

    execute_invocation: Union[FunctionInvocation, RevertedFunctionInvocation]
    execution_resources: ExecutionResources
    validate_invocation: Optional[FunctionInvocation] = None
    fee_transfer_invocation: Optional[FunctionInvocation] = None
    state_diff: Optional[StateDiff] = None


@dataclass
class DeclareTransactionTrace:
    """
    Dataclass representing a transaction trace of an DECLARE transaction.
    """

    execution_resources: ExecutionResources
    validate_invocation: Optional[FunctionInvocation] = None
    fee_transfer_invocation: Optional[FunctionInvocation] = None
    state_diff: Optional[StateDiff] = None


@dataclass
class DeployAccountTransactionTrace:
    """
    Dataclass representing a transaction trace of an DEPLOY_ACCOUNT transaction.
    """

    constructor_invocation: FunctionInvocation
    execution_resources: ExecutionResources
    validate_invocation: Optional[FunctionInvocation] = None
    fee_transfer_invocation: Optional[FunctionInvocation] = None
    state_diff: Optional[StateDiff] = None


@dataclass
class L1HandlerTransactionTrace:
    """
    Dataclass representing a transaction trace of an L1_HANDLER transaction.
    """

    execution_resources: ExecutionResources
    function_invocation: FunctionInvocation
    state_diff: Optional[StateDiff] = None


TransactionTrace = Union[
    InvokeTransactionTrace,
    DeclareTransactionTrace,
    DeployAccountTransactionTrace,
    L1HandlerTransactionTrace,
]


@dataclass
class SimulatedTransaction:
    """
    Dataclass representing a simulated transaction returned by `starknet_simulateTransactions` method.
    """

    transaction_trace: TransactionTrace
    fee_estimation: EstimatedFee


@dataclass
class BlockTransactionTrace:
    """
    Dataclass representing a single transaction trace in a block.
    """

    transaction_hash: int
    trace_root: TransactionTrace


@dataclass
class BinaryNode:
    """
    Dataclass representing an internal node whose both children are non-zero.
    """

    left: int
    right: int


@dataclass
class EdgeNode:
    """
    Dataclass representing a path to the highest non-zero descendant node.
    """

    path: int
    length: int
    child: int


MerkleNode = Union[BinaryNode, EdgeNode]


@dataclass
class NodeHashToNodeMappingItem:
    node_hash: int
    node: MerkleNode


NodeHashToNodeMapping = List[NodeHashToNodeMappingItem]


@dataclass
class ContractsStorageKeys:
    """
    Dataclass representing a pair of contract address and storage keys.
    """

    contract_address: int
    storage_keys: List[str]


@dataclass
class ContractLeafData:
    nonce: int
    class_hash: int
    storage_root: Optional[int]


@dataclass
class GlobalRoots:
    contracts_tree_root: int
    classes_tree_root: int
    block_hash: int


@dataclass
class ContractsProof:
    nodes: NodeHashToNodeMapping
    contract_leaves_data: List[ContractLeafData]


@dataclass
class StorageProofResponse:
    """
    Dataclass representing a response to a storage proof request.
    """

    classes_proof: NodeHashToNodeMapping
    contracts_proof: ContractsProof
    contracts_storage_proofs: List[NodeHashToNodeMapping]
    global_roots: GlobalRoots


@dataclass
class MessageStatus:
    transaction_hash: int
    finality_status: TransactionFinalityStatus
    execution_status: TransactionExecutionStatus
    failure_reason: Optional[str] = None


@dataclass
class OutsideExecution:
    """
    Dataclass representing an outside execution.
    (SNIP-9)[https://github.com/starknet-io/SNIPs/blob/main/SNIPS/snip-9.md]
    """

    caller: int
    nonce: int
    execute_after: int
    execute_before: int
    calls: List[Call]

    # TODO(#1537): Use serialiser to convert to ABI dict.
    def to_abi_dict(self) -> Dict:
        """
        Returns a dictionary that can be serialized (compiled) into calldata
        using StructSerializer
        """
        return {
            "caller": self.caller,
            "nonce": self.nonce,
            "execute_after": self.execute_after,
            "execute_before": self.execute_before,
            "calls": [
                {
                    "to": call.to_addr,
                    "selector": call.selector,
                    "calldata": call.calldata,
                }
                for call in self.calls
            ],
        }


@dataclass
class _EmittedEventWithFinalityStatus:
    finality_status: TransactionFinalityStatus


@dataclass
class EmittedEventWithFinalityStatus(
    _EmittedEventDefaultBase, _EmittedEventWithFinalityStatus, _EmittedEventBase
):
    """
    Dataclass representing an event emitted by transaction.
    """



from typing import List, Optional, Sequence, Tuple

from semver import Version

from starknet_py.cairo.felt import encode_shortstring
from starknet_py.hash.compiled_class_hash_objects import (
    BytecodeLeaf,
    BytecodeSegment,
    BytecodeSegmentedNode,
    BytecodeSegmentStructure,
    NestedIntList,
)
from starknet_py.hash.hash_method import HashMethod
from starknet_py.net.client_models import CasmClassEntryPoint
from starknet_py.net.executable_models import CasmClass

CASM_CLASS_VERSION = "COMPILED_CLASS_V1"


def get_casm_hash_method_for_starknet_version(starknet_version: Version) -> HashMethod:
    # Starknet 0.14.1 and later use Blake2s
    if starknet_version >= Version.parse("0.14.1"):
        return HashMethod.BLAKE2S

    return HashMethod.POSEIDON


def compute_casm_class_hash(
    casm_contract_class: CasmClass, hash_method: HashMethod = HashMethod.POSEIDON
) -> int:
    """
    Calculate class hash of a CasmClass.
    """
    casm_class_version = encode_shortstring(CASM_CLASS_VERSION)

    _entry_points = casm_contract_class.entry_points_by_type

    external_entry_points_hash = hash_method.hash_many(
        _entry_points_array(_entry_points.external, hash_method)
    )
    l1_handler_entry_points_hash = hash_method.hash_many(
        _entry_points_array(_entry_points.l1_handler, hash_method)
    )
    constructor_entry_points_hash = hash_method.hash_many(
        _entry_points_array(_entry_points.constructor, hash_method)
    )

    if casm_contract_class.bytecode_segment_lengths is not None:
        bytecode_hash = create_bytecode_segment_structure(
            bytecode=casm_contract_class.bytecode,
            bytecode_segment_lengths=casm_contract_class.bytecode_segment_lengths,
            visited_pcs=None,
        ).hash(hash_method)
    else:
        bytecode_hash = hash_method.hash_many(casm_contract_class.bytecode)

    return hash_method.hash_many(
        [
            casm_class_version,
            external_entry_points_hash,
            l1_handler_entry_points_hash,
            constructor_entry_points_hash,
            bytecode_hash,
        ]
    )


def _entry_points_array(
    entry_points: List[CasmClassEntryPoint], hash_method: HashMethod
) -> List[int]:
    entry_points_array = []
    for entry_point in entry_points:
        assert entry_point.builtins is not None
        _encoded_builtins = [encode_shortstring(val) for val in entry_point.builtins]
        builtins_hash = hash_method.hash_many(_encoded_builtins)

        entry_points_array.extend(
            [entry_point.selector, entry_point.offset, builtins_hash]
        )

    return entry_points_array


# create_bytecode_segment_structure and _create_bytecode_segment_structure_inner are copied from
# https://github.com/starkware-libs/cairo-lang/blob/v0.13.1/src/starkware/starknet/core/os/contract_class/compiled_class_hash.py


def create_bytecode_segment_structure(
    bytecode: List[int],
    bytecode_segment_lengths: NestedIntList,
    visited_pcs: Optional[Sequence[int]],
) -> BytecodeSegmentStructure:
    """
    Creates a BytecodeSegmentStructure instance from the given bytecode and
    bytecode_segment_lengths.
    """
    rev_visited_pcs = list(
        visited_pcs if visited_pcs is not None else range(len(bytecode))
    )[::-1]

    res, total_len = _create_bytecode_segment_structure_inner(
        bytecode=bytecode,
        bytecode_segment_lengths=bytecode_segment_lengths,
        visited_pcs=rev_visited_pcs,
        bytecode_offset=0,
    )
    assert total_len == len(
        bytecode
    ), f"Invalid length bytecode segment structure: {total_len}. Bytecode length: {len(bytecode)}."
    assert len(rev_visited_pcs) == 0, f"PC {rev_visited_pcs[-1]} is out of range."
    return res


def _create_bytecode_segment_structure_inner(
    bytecode: List[int],
    bytecode_segment_lengths: NestedIntList,
    visited_pcs: List[int],
    bytecode_offset: int,
) -> Tuple[BytecodeSegmentStructure, int]:
    """
    Helper function for `create_bytecode_segment_structure`.
    `visited_pcs` should be given in reverse order, and is consumed by the function.
    Returns the BytecodeSegmentStructure and the total length of the processed segment.
    """
    if isinstance(bytecode_segment_lengths, int):
        segment_end = bytecode_offset + bytecode_segment_lengths

        # Remove all the visited PCs that are in the segment.
        while len(visited_pcs) > 0 and bytecode_offset <= visited_pcs[-1] < segment_end:
            visited_pcs.pop()

        return (
            BytecodeLeaf(data=bytecode[bytecode_offset:segment_end]),
            bytecode_segment_lengths,
        )

    res = []
    total_len = 0
    for item in bytecode_segment_lengths:
        visited_pc_before = visited_pcs[-1] if len(visited_pcs) > 0 else None

        current_structure, item_len = _create_bytecode_segment_structure_inner(
            bytecode=bytecode,
            bytecode_segment_lengths=item,
            visited_pcs=visited_pcs,
            bytecode_offset=bytecode_offset,
        )

        visited_pc_after = visited_pcs[-1] if len(visited_pcs) > 0 else None
        is_used = visited_pc_after != visited_pc_before

        if is_used and visited_pc_before != bytecode_offset:
            raise ValueError(
                f"Invalid segment structure: PC {visited_pc_before} was visited, "
                f"but the beginning of the segment ({bytecode_offset}) was not."
            )

        res.append(
            BytecodeSegment(
                segment_length=item_len,
                is_used=is_used,
                inner_structure=current_structure,
            )
        )
        bytecode_offset += item_len
        total_len += item_len

    return BytecodeSegmentedNode(segments=res), total_len



import pytest

from starknet_py.serialization.data_serializers.option_serializer import (
    OptionSerializer,
)
from starknet_py.serialization.data_serializers.uint_serializer import UintSerializer


@pytest.mark.parametrize(
    "serializer, value, serialized_value",
    [
        (OptionSerializer(UintSerializer(128)), 123, [0, 123]),
        (OptionSerializer(UintSerializer(256)), 1, [0, 1, 0]),
        (OptionSerializer(UintSerializer(128)), None, [1]),
        (OptionSerializer(UintSerializer(256)), None, [1]),
    ],
)
def test_option_serializer(serializer, value, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value

    serialized = serializer.serialize(value)
    assert serialized == serialized_value



import re
from collections import OrderedDict

import pytest

from starknet_py.serialization.data_serializers.option_serializer import (
    OptionSerializer,
)
from starknet_py.serialization.data_serializers.output_serializer import (
    OutputSerializer,
)
from starknet_py.serialization.data_serializers.struct_serializer import (
    StructSerializer,
)
from starknet_py.serialization.data_serializers.uint_serializer import UintSerializer
from starknet_py.tests.unit.serialization.data_serializers.uint256_serializer_test import (
    SHIFT,
)

serializer = OutputSerializer(
    serializers=[
        UintSerializer(256),
        OptionSerializer(
            StructSerializer(
                OrderedDict(
                    my_option=OptionSerializer(UintSerializer(128)),
                    my_uint=UintSerializer(256),
                )
            )
        ),
    ]
)


@pytest.mark.parametrize(
    "value, serialized_value",
    [
        (
            (1 + 1 * SHIFT, OrderedDict(my_option=123, my_uint=1 + 1 * SHIFT)),
            [1, 1, 0, 0, 123, 1, 1],
        ),
        ((0, OrderedDict(my_option=None, my_uint=1)), [0, 0, 0, 1, 1, 0]),
        ((1, None), [1, 0, 1]),
    ],
)
def test_output_serializer_deserialize(value, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value


def test_output_serializer_serialize():
    error_message = re.escape(
        "Output serializer can't be used to transform python data into calldata."
    )
    with pytest.raises(ValueError, match=error_message):
        serializer.serialize([1, None])



from dataclasses import dataclass
from secrets import token_bytes

from eth_keyfile.keyfile import extract_key_from_keyfile

from starknet_py.constants import FIELD_PRIME
from starknet_py.hash.utils import private_to_stark_key
from starknet_py.net.client_models import Hash


@dataclass
class KeyPair:
    private_key: int
    public_key: int

    def __init__(self, private_key: Hash, public_key: Hash):
        if isinstance(private_key, str):
            self.private_key = int(private_key, 0)
        else:
            self.private_key = private_key

        if isinstance(public_key, str):
            self.public_key = int(public_key, 0)
        else:
            self.public_key = public_key

    @staticmethod
    def generate() -> "KeyPair":
        """
        Create a key pair from a randomly generated private key.

        :return: KeyPair object.
        """
        random_int = int.from_bytes(token_bytes(32), byteorder="big")
        private_key = random_int % FIELD_PRIME
        public_key = private_to_stark_key(private_key)
        return KeyPair(private_key=private_key, public_key=public_key)

    @staticmethod
    def from_private_key(key: Hash) -> "KeyPair":
        if isinstance(key, str):
            key = int(key, 0)
        return KeyPair(private_key=key, public_key=private_to_stark_key(key))

    @staticmethod
    def from_keystore(path: str, password: str) -> "KeyPair":
        """
        Create a key pair from a keystore file.
        The keystore file should follow the Ethereum keystore format.

        :param path: Path to the keystore file.
        :param password: Password to decrypt the keystore file.
        :return: KeyPair object.
        """
        key = extract_key_from_keyfile(path, password)
        return KeyPair.from_private_key(int.from_bytes(key, byteorder="big"))



import dataclasses
import re
from unittest.mock import AsyncMock, Mock, patch

import pytest

from starknet_py.common import create_sierra_compiled_contract
from starknet_py.contract import Contract, DeclareResult
from starknet_py.net.client_models import InvokeTransactionV3
from starknet_py.net.models import DeclareV3
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import load_contract


@pytest.mark.asyncio
async def test_declare_deploy_v3(
    account,
    minimal_contract_class_hash: int,
):
    compiled_contract = load_contract("MinimalContract")["sierra"]

    declare_result = DeclareResult(
        _account=account,
        _client=account.client,
        _cairo_version=1,
        class_hash=minimal_contract_class_hash,
        compiled_contract=compiled_contract,
        hash=0,
        declare_transaction=Mock(spec=DeclareV3),
    )

    tip = 12345
    deploy_result = await declare_result.deploy_v3(
        resource_bounds=MAX_RESOURCE_BOUNDS, tip=tip
    )
    await deploy_result.wait_for_acceptance()

    assert isinstance(deploy_result.hash, int)
    assert deploy_result.hash != 0
    assert deploy_result.deployed_contract.address != 0
    transaction = await account.client.get_transaction(deploy_result.hash)
    assert isinstance(transaction, InvokeTransactionV3)
    assert transaction.tip == tip


@pytest.mark.asyncio
async def test_declare_deploy_v3_auto_estimate_tip(
    account,
    minimal_contract_class_hash: int,
    get_block_with_txs_path,
    block_with_tips_mock,
):
    compiled_contract = load_contract("MinimalContract")["sierra"]

    declare_result = DeclareResult(
        _account=account,
        _client=account.client,
        _cairo_version=1,
        class_hash=minimal_contract_class_hash,
        compiled_contract=compiled_contract,
        hash=0,
        declare_transaction=Mock(spec=DeclareV3),
    )

    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        get_block_with_txs_mock.return_value = block_with_tips_mock

        deploy_result = await declare_result.deploy_v3(
            resource_bounds=MAX_RESOURCE_BOUNDS, auto_estimate_tip=True
        )

    await deploy_result.wait_for_acceptance()

    assert isinstance(deploy_result.hash, int)
    assert deploy_result.hash != 0
    assert deploy_result.deployed_contract.address != 0
    transaction = await account.client.get_transaction(deploy_result.hash)
    assert isinstance(transaction, InvokeTransactionV3)
    assert transaction.tip == 2


@pytest.mark.asyncio
async def test_throws_on_wrong_abi(account, minimal_contract_class_hash: int):
    compiled_contract = load_contract("MinimalContract")["sierra"]

    declare_result = DeclareResult(
        _account=account,
        _client=account.client,
        _cairo_version=1,
        class_hash=minimal_contract_class_hash,
        compiled_contract=compiled_contract,
        hash=0,
        declare_transaction=Mock(spec=DeclareV3),
    )

    compiled_contract = compiled_contract.replace('"abi":[', '"api": ')

    declare_result = dataclasses.replace(
        declare_result, compiled_contract=compiled_contract
    )
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Contract's ABI can't be converted to format List[Dict]. "
            "Make sure provided compiled_contract is correct."
        ),
    ):
        await declare_result.deploy_v3(resource_bounds=MAX_RESOURCE_BOUNDS)


@pytest.mark.asyncio
async def test_deploy_contract_v3(account, hello_starknet_class_hash: int):
    compiled_contract = load_contract("HelloStarknet")["sierra"]
    abi = create_sierra_compiled_contract(
        compiled_contract=compiled_contract
    ).parsed_abi

    tip = 12345
    deploy_result = await Contract.deploy_contract_v3(
        class_hash=hello_starknet_class_hash,
        account=account,
        abi=abi,
        resource_bounds=MAX_RESOURCE_BOUNDS,
        tip=tip,
    )
    await deploy_result.wait_for_acceptance()

    contract = deploy_result.deployed_contract

    assert isinstance(contract.address, int)
    assert len(contract.functions) != 0

    transaction = await account.client.get_transaction(tx_hash=deploy_result.hash)
    assert isinstance(transaction, InvokeTransactionV3)
    assert transaction.tip == tip

    class_hash = await account.client.get_class_hash_at(
        contract_address=contract.address
    )
    assert class_hash == hello_starknet_class_hash



# fmt: off
import pytest
from semver import Version

from starknet_py.common import create_casm_class
from starknet_py.hash.casm_class_hash import (
    compute_casm_class_hash,
    get_casm_hash_method_for_starknet_version,
)
from starknet_py.hash.hash_method import HashMethod
from starknet_py.tests.e2e.fixtures.constants import PRECOMPILED_CONTRACTS_DIR
from starknet_py.tests.e2e.fixtures.misc import (
    ContractVersion,
    load_contract,
    read_contract,
)


@pytest.mark.parametrize(
    "contract, expected_casm_class_hash_poseidon",
    [
        ("Account", 0x778bce178afd1b39abd9729b80931e8c71661103b16de928c3187057254f601),
        ("ERC20", 0x3748ca8b6c53d65b5862e6f17850033baa117075e887708474aba110cc0e77a),
        ("HelloStarknet", 0x467aa3aa331f2d1f9ca24168ad6e9836685af53c35c53a4a828f3564efe79cd),
        ("TestContract", 0x5ab6a30579a54901b6e8be599d7cbdf00a1e05445e95e1d429ca0d70b408af7),
        ("TokenBridge", 0xf364f0d735b07f5a9a50a886e1f5bf6f0d82175d1955dc737f998d33990f8e),
    ],
)
def test_compute_casm_class_hash_with_poseidon(contract, expected_casm_class_hash_poseidon):
    casm_contract_class_str = load_contract(
        contract, version=ContractVersion.V2
    )['casm']

    casm_class = create_casm_class(casm_contract_class_str)
    casm_class_hash = compute_casm_class_hash(casm_class)
    assert casm_class_hash == expected_casm_class_hash_poseidon

@pytest.mark.parametrize(
    "casm_contract_class_source, expected_casm_class_hash_poseidon",
    [
        ("minimal_contract_compiled_v2_1.casm",
         0x186f6c4ca3af40dbcbf3f08f828ab0ee072938aaaedccc74ef3b9840cbd9fb3),
        ("minimal_contract_compiled_v2_5_4.casm",
         0x1d055a90aa90db474fa08a931d5e63753c6f762fa3e9597b26c8d4b003a2de6),
        ("starknet_contract_v2_6.casm", 0x603dd72504d8b0bc54df4f1102fdcf87fc3b2b94750a9083a5876913eec08e4),
    ],
)
def test_precompiled_compute_casm_class_hash_with_poseidon(casm_contract_class_source, expected_casm_class_hash_poseidon): # pylint: disable=line-too-long
    casm_contract_class_str = read_contract(
        casm_contract_class_source, directory=PRECOMPILED_CONTRACTS_DIR
    )

    casm_class = create_casm_class(casm_contract_class_str)
    casm_class_hash = compute_casm_class_hash(casm_class)
    assert casm_class_hash == expected_casm_class_hash_poseidon


@pytest.mark.parametrize(
    "starknet_version, expected_hash_method",
    [
        ("0.13.5", HashMethod.POSEIDON),
        ("0.14.0", HashMethod.POSEIDON),
        ("0.14.1", HashMethod.BLAKE2S),
        ("0.15.0", HashMethod.BLAKE2S),
        ("1.0.0", HashMethod.BLAKE2S),
        ("1.10.0", HashMethod.BLAKE2S),
    ],
)
def test_get_casm_hash_method_for_starknet_version(starknet_version, expected_hash_method):
    """Test that the correct hash method is returned for different Starknet versions."""
    starknet_version = Version.parse(starknet_version)
    hash_method = get_casm_hash_method_for_starknet_version(starknet_version)
    assert hash_method == expected_hash_method


@pytest.mark.parametrize(
    "contract, expected_casm_class_hash_blake2s",
    [
        ("Account", 0x714c833f7b359955f6a4a495ba995cca2114158db2178aff587f643daa19c80),
        ("ERC20", 0x44312efaec9c719168eee3586314b01ed7a1fd7e31d3cf0c5a17e0a5b4fbe7d),
        ("HelloStarknet", 0x5aaedd0566b5dd234f5f8d3d6b8cfd299cf0a99541aa9ca34db9259d546e82f),
        ("TestContract", 0x3135acde04efbc96d422c01822a517ae5b4e61f132d26bf8542e3b9d0d1500f),
        ("TokenBridge", 0x6409448fd244060b15748b02b6e0bdb185d5271be231492ca33a7147e43994c),
    ],
)

def test_compute_casm_class_hash_with_blake2s(contract, expected_casm_class_hash_blake2s):
    casm_contract_class_str = load_contract(
        contract, version=ContractVersion.V2
    )['casm']

    casm_class = create_casm_class(casm_contract_class_str)
    casm_class_hash = compute_casm_class_hash(casm_class, hash_method=HashMethod.BLAKE2S)
    assert casm_class_hash == expected_casm_class_hash_blake2s

@pytest.mark.parametrize(
    "casm_contract_class_source, expected_casm_class_hash_blake2s",
    [
        ("minimal_contract_compiled_v2_1.casm",
         0x195cfeec43b384e0f0ec83937149a1a4d88571772b2806ed7e4f41a1ecb4c74),
        ("minimal_contract_compiled_v2_5_4.casm",
         0x5ac03c50c46fc7b374d4e11d15693ae0d21e13f61c1704700294d1f378980f7),
        ("starknet_contract_v2_6.casm", 0xf8c27dd667e50ba127e5e0e469381606ffece27d8c5148548b6bbc4cacf717),
    ],
)
def test_precompiled_compute_casm_class_hash_with_blake2s(casm_contract_class_source, expected_casm_class_hash_blake2s):
    casm_contract_class_str = read_contract(
        casm_contract_class_source, directory=PRECOMPILED_CONTRACTS_DIR
    )

    casm_class = create_casm_class(casm_contract_class_str)
    casm_class_hash = compute_casm_class_hash(casm_class, hash_method=HashMethod.BLAKE2S)
    assert casm_class_hash == expected_casm_class_hash_blake2s



from marshmallow import fields
from marshmallow_oneofschema.one_of_schema import OneOfSchema

from starknet_py.net.client_models import TransactionType
from starknet_py.net.schemas.rpc.contract import SierraCompiledContractSchema
from starknet_py.net.schemas.rpc.transactions import (
    DeclareTransactionV3Schema,
    DeployAccountTransactionV3Schema,
    InvokeTransactionV3Schema,
)


class BroadcastedDeclareV3Schema(DeclareTransactionV3Schema):
    contract_class = fields.Nested(
        SierraCompiledContractSchema(), data_key="contract_class", required=True
    )


class BroadcastedTransactionSchema(OneOfSchema):
    type_schemas = {
        TransactionType.INVOKE.name: InvokeTransactionV3Schema(),
        TransactionType.DECLARE.name: BroadcastedDeclareV3Schema(),
        TransactionType.DEPLOY_ACCOUNT.name: DeployAccountTransactionV3Schema(),
    }

    def get_obj_type(self, obj):
        return obj.type.name



from __future__ import annotations

from collections import OrderedDict
from typing import Dict, List, Union

from starknet_py.abi.v0 import Abi as AbiV0
from starknet_py.abi.v1 import Abi as AbiV1
from starknet_py.abi.v2 import Abi as AbiV2
from starknet_py.cairo.data_types import (
    ArrayType,
    BoolType,
    CairoType,
    EnumType,
    EventType,
    FeltType,
    FixedSizeArrayType,
    IntType,
    NamedTupleType,
    NonZeroType,
    OptionType,
    StructType,
    TupleType,
    UintType,
    UnitType,
)
from starknet_py.serialization.data_serializers import (
    BoolSerializer,
    ByteArraySerializer,
)
from starknet_py.serialization.data_serializers.array_serializer import ArraySerializer
from starknet_py.serialization.data_serializers.cairo_data_serializer import (
    CairoDataSerializer,
)
from starknet_py.serialization.data_serializers.enum_serializer import EnumSerializer
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.data_serializers.int_serializer import IntSerializer
from starknet_py.serialization.data_serializers.named_tuple_serializer import (
    NamedTupleSerializer,
)
from starknet_py.serialization.data_serializers.non_zero_serializer import (
    NonZeroSerializer,
)
from starknet_py.serialization.data_serializers.option_serializer import (
    OptionSerializer,
)
from starknet_py.serialization.data_serializers.output_serializer import (
    OutputSerializer,
)
from starknet_py.serialization.data_serializers.payload_serializer import (
    PayloadSerializer,
)
from starknet_py.serialization.data_serializers.struct_serializer import (
    StructSerializer,
)
from starknet_py.serialization.data_serializers.tuple_serializer import TupleSerializer
from starknet_py.serialization.data_serializers.uint256_serializer import (
    Uint256Serializer,
)
from starknet_py.serialization.data_serializers.uint_serializer import UintSerializer
from starknet_py.serialization.data_serializers.unit_serializer import UnitSerializer
from starknet_py.serialization.errors import InvalidTypeException
from starknet_py.serialization.function_serialization_adapter import (
    FunctionSerializationAdapterV0,
    FunctionSerializationAdapterV1,
)

_uint256_type = StructType("Uint256", OrderedDict(low=FeltType(), high=FeltType()))
_byte_array_type = StructType(
    "core::byte_array::ByteArray",
    OrderedDict(
        data=ArrayType(FeltType()),
        pending_word=FeltType(),
        pending_word_len=UintType(bits=32),
    ),
)


def serializer_for_type(cairo_type: CairoType) -> CairoDataSerializer:
    """
    Create a serializer for cairo type.

    :param cairo_type: CairoType.
    :return: CairoDataSerializer.
    """
    # pylint: disable=too-many-return-statements, too-many-branches
    if isinstance(cairo_type, FeltType):
        return FeltSerializer()

    if isinstance(cairo_type, BoolType):
        return BoolSerializer()

    if isinstance(cairo_type, StructType):
        # Special case: Uint256 is represented as struct
        if cairo_type == _uint256_type:
            return Uint256Serializer()

        if cairo_type == _byte_array_type:
            return ByteArraySerializer()

        return StructSerializer(
            OrderedDict(
                (name, serializer_for_type(member_type))
                for name, member_type in cairo_type.types.items()
            )
        )

    if isinstance(cairo_type, (ArrayType, FixedSizeArrayType)):
        return ArraySerializer(serializer_for_type(cairo_type.inner_type))

    if isinstance(cairo_type, TupleType):
        return TupleSerializer(
            [serializer_for_type(member) for member in cairo_type.types]
        )

    if isinstance(cairo_type, NamedTupleType):
        return NamedTupleSerializer(
            OrderedDict(
                (name, serializer_for_type(member_type))
                for name, member_type in cairo_type.types.items()
            )
        )

    if isinstance(cairo_type, IntType):
        return IntSerializer(bits=cairo_type.bits)

    if isinstance(cairo_type, UintType):
        return UintSerializer(bits=cairo_type.bits)

    if isinstance(cairo_type, OptionType):
        return OptionSerializer(serializer_for_type(cairo_type.type))

    if isinstance(cairo_type, NonZeroType):
        return NonZeroSerializer(serializer_for_type(cairo_type.type))

    if isinstance(cairo_type, UnitType):
        return UnitSerializer()

    if isinstance(cairo_type, EnumType):
        return EnumSerializer(
            OrderedDict(
                (name, serializer_for_type(variant_type))
                for name, variant_type in cairo_type.variants.items()
            )
        )
    if isinstance(cairo_type, EventType):
        return serializer_for_payload(cairo_type.types)

    raise InvalidTypeException(f"Received unknown Cairo type '{cairo_type}'.")


# We don't want to require users to use OrderedDict. Regular python requires order since python 3.7.
def serializer_for_payload(payload: Dict[str, CairoType]) -> PayloadSerializer:
    """
    Create PayloadSerializer for types listed in a dictionary. Please note that the order of fields in the dict is
    very important. Make sure the keys are provided in the right order.

    :param payload: dictionary with cairo types.
    :return: PayloadSerializer that can be used to (de)serialize events/function calls.
    """
    return PayloadSerializer(
        OrderedDict(
            (name, serializer_for_type(cairo_type))
            for name, cairo_type in payload.items()
        )
    )


def serializer_for_outputs(payload: List[CairoType]) -> OutputSerializer:
    """
    Create OutputSerializer for types in list. Please note that the order of fields in the list is
    very important. Make sure the types are provided in the right order.

    :param payload: list with cairo types.
    :return: OutputSerializer that can be used to deserialize function outputs.
    """
    return OutputSerializer(
        serializers=[serializer_for_type(cairo_type) for cairo_type in payload]
    )


EventV0 = AbiV0.Event
EventV1 = AbiV1.Event
EventV2 = EventType


def serializer_for_event(event: EventV0 | EventV1 | EventV2) -> PayloadSerializer:
    """
    Create serializer for an event.

    :param event: parsed event.
    :return: PayloadSerializer that can be used to (de)serialize events.
    """
    if isinstance(event, EventV0):
        return serializer_for_payload(event.data)
    if isinstance(event, EventV1):
        return serializer_for_payload(event.inputs)
    return serializer_for_payload(event.types)


def serializer_for_function(
    abi_function: AbiV0.Function,
) -> FunctionSerializationAdapterV0:
    """
    Create FunctionSerializationAdapter for serializing function inputs and deserializing function outputs.

    :param abi_function: parsed function's abi.
    :return: FunctionSerializationAdapter.
    """
    return FunctionSerializationAdapterV0(
        inputs_serializer=serializer_for_payload(abi_function.inputs),
        outputs_deserializer=serializer_for_payload(abi_function.outputs),
    )


def serializer_for_function_v1(
    abi_function: Union[AbiV1.Function, AbiV2.Function],
) -> FunctionSerializationAdapterV1:
    """
    Create FunctionSerializationAdapter for serializing function inputs and deserializing function outputs.

    :param abi_function: parsed function's abi.
    :return: FunctionSerializationAdapter.
    """
    return FunctionSerializationAdapterV1(
        inputs_serializer=serializer_for_payload(abi_function.inputs),
        outputs_deserializer=serializer_for_outputs(abi_function.outputs),
    )


def serializer_for_constructor_v2(
    abi_function: AbiV2.Constructor,
) -> FunctionSerializationAdapterV1:
    """
    Create FunctionSerializationAdapter for serializing constructor inputs.

    :param abi_function: parsed constructor's abi.
    :return: FunctionSerializationAdapter.
    """
    return FunctionSerializationAdapterV1(
        inputs_serializer=serializer_for_payload(abi_function.inputs),
        outputs_deserializer=serializer_for_outputs([]),
    )



import pytest

from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_multicall(account, deployed_balance_contract):
    # pylint: disable=import-outside-toplevel
    balance_contract = deployed_balance_contract
    (initial_balance,) = await balance_contract.functions["get_balance"].call()
    # docs: start
    from starknet_py.hash.selector import get_selector_from_name
    from starknet_py.net.client_models import Call

    increase_balance_by_20_call = Call(
        to_addr=balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[20],
    )
    calls = [increase_balance_by_20_call, increase_balance_by_20_call]

    # Execute one transaction with multiple calls
    resp = await account.execute_v3(
        calls=calls,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await account.client.wait_for_tx(resp.transaction_hash)
    # docs: end

    (final_balance,) = await balance_contract.functions["get_balance"].call()
    assert final_balance == initial_balance + 40
